searchData={"items":[{"type":"module","title":"Sourceror","doc":"Utilities to work with Elixir source code.","ref":"Sourceror.html"},{"type":"module","title":"Installation - Sourceror","doc":"Add `:sourceror` as a dependency to your project's `mix.exs`:\n\n```elixir\ndefp deps do\n  [\n    {:sourceror, \"~> 1.3\"}\n  ]\nend\n```\n\n#","ref":"Sourceror.html#module-installation"},{"type":"module","title":"A note on compatibility - Sourceror","doc":"Sourceror is compatible with Elixir versions down to 1.10 and OTP 21. For Elixir\nversions prior to 1.13 it uses a vendored version of the Elixir parser and\nformatter modules. This means that for Elixir versions prior to 1.12 it will\nsuccessfully parse the new syntax for stepped ranges instead of raising a\n`SyntaxError`, but everything else should work as expected.","ref":"Sourceror.html#module-a-note-on-compatibility"},{"type":"module","title":"Goals of the library - Sourceror","doc":"- Be as close as possible to the standard Elixir AST.\n- Make working with comments as simple as possible.\n- No dev/prod dependencies, to simplify integration with other tools.","ref":"Sourceror.html#module-goals-of-the-library"},{"type":"module","title":"Sourceror's AST - Sourceror","doc":"Having the AST and comments as separate entities allows Elixir to expose the\ncode formatting utilities without making any changes to it's AST, but also\ndelegates the task of figuring out what's the most appropriate way to work with\nthem to us.\n\nSourceror's take is to use the node metadata to store the comments. This allows\nus to work with an AST that is as close to regular elixir AST as possible. It\nalso allows you to move nodes around without worrying about leaving a comment\nbehind and ending up with misplaced comments.\n\nTwo metadata fields are added to the regular Elixir AST:\n\n- `:leading_comments` - holds the comments directly above the node or are in the\n  same line as it. For example:\n\n  ```elixir\n  test \"parses leading comments\" do\n    quoted = \"\"\"\n    # Comment for :a\n    :a # Also a comment for :a\n    \"\"\" |> Sourceror.parse_string!()\n\n    assert {:__block__, meta, [:a]} = quoted\n    assert meta[:leading_comments] == [\n      %{line: 1, column: 1, previous_eol_count: 1, next_eol_count: 1, text: \"# Comment for :a\"},\n      %{line: 2, column: 4, previous_eol_count: 0, next_eol_count: 1, text: \"# Also a comment for :a\"},\n    ]\n  end\n  ```\n\n- `:trailing_comments` - holds the comments that are inside of the node, but\n  aren't leading any children, for example:\n\n  ```elixir\n  test \"parses trailing comments\" do\n    quoted = \"\"\"\n    def foo() do\n    :ok\n    # A trailing comment\n    end # Not a trailing comment for :foo\n    \"\"\" |> Sourceror.parse_string!()\n\n    assert {:__block__, block_meta, [{:def, meta, _}]} = quoted\n    assert [%{line: 3, text: \"# A trailing comment\"}] = meta[:trailing_comments]\n    assert [%{line: 4, text: \"# Not a trailing comment for :foo\"}] = block_meta[:trailing_comments]\n  end\n  ```\n\nNote that Sourceror considers leading comments to the ones that are found in the\nsame line as a node, and trailing comments to the ones that are found before the\nending line of a node, based on the `end`, `closing` or `end_of_expression`\nline. This also makes the Sourceror AST consistent with the way the Elixir\nformatter works, making it easier to reason about how a given AST would be\nformatted.","ref":"Sourceror.html#module-sourceror-s-ast"},{"type":"module","title":"Traversing the AST - Sourceror","doc":"Elixir provides the `Macro.prewalk`, `Macro.postwalk` and `Macro.traverse`\nfunctions to traverse the AST. You can use the same functions to traverse the\nSourceror AST as well, since it has the same shape as the standard Elixir AST.\n\nSourceror also provides the `Sourceror.prewalk`, `Sourceror.postwalk` and\n`Sourceror.traverse` variants. At the time of writing they are mostly wrappers\naround the standard Elixir functions for AST traversal, but they may be enhanced\nin the future if more AST formats are introduced.\n\nIn addition to these, Sourceror also provides a Zipper implementation for the\nElixir AST. You can learn more about it in the [Zippers\nnotebook](https://hexdocs.pm/sourceror/zippers.html).","ref":"Sourceror.html#module-traversing-the-ast"},{"type":"module","title":"Patching the source code - Sourceror","doc":"You can use Sourceror to manipulate the AST and turn it back into human readable\nElixir code, this is commonly known as writing a \"codemod\". For example, you can\nwrite a codemod to replace calls to `String.to_atom` to\n`String.to_existing_atom`:\n\n```elixir\ntest \"updates the source code\" do\n  source =\n    \"\"\"\n    String.to_atom(foo)\\\n    \"\"\"\n\n  new_source =\n    source\n    |> Sourceror.parse_string!()\n    |> Macro.postwalk(fn\n      {{:., dot_meta, [{:__aliases__, alias_meta, [:String]}, :to_atom]}, call_meta, args} ->\n        {{:., dot_meta, [{:__aliases__, alias_meta, [:String]}, :to_existing_atom]}, call_meta, args}\n\n      quoted ->\n        quoted\n    end)\n    |> Sourceror.to_string()\n\n  assert new_source ==\n    \"\"\"\n    String.to_existing_atom(foo)\\\n    \"\"\"\nend\n```\n\nHowever, this will affect the whole source code, as we are working on the full\nsource AST. Sourceror relies on the Elixir formatter to produce human readable\ncode, so the original code formatting will be lost by using\n`Sourceror.to_string`. If your code is already using the Elixir formatter then\nthis won't be an issue, but it will be an undesirable effect if you're not using\nit.\n\nAn alternative to this is to use Patches instead. A patch is a data structure\nthat specifies the text range that should be replaced, and either a replacement\nstring or a function that takes the text in that range and produces a string\nreplacement.\n\nUsing patches, we could do the same as above, but produce a patch instead of\nmodifying the AST. As a result, only the parts that need to be changed will be\naffected, and the rest of the code keeps the original formatting:\n\n```elixir\ntest \"patches the source code\" do\n  source =\n    \"\"\"\n    case foo do\n      nil ->         :bar\n      _ ->\n\n          String.to_atom(foo)\n\n          end\\\n    \"\"\"\n\n  {_quoted, patches} =\n    source\n    |> Sourceror.parse_string!()\n    |> Macro.postwalk([], fn\n      {{:., dot_meta, [{:__aliases__, alias_meta, [:String]}, :to_atom]}, call_meta, args} = quoted, patches ->\n        range = Sourceror.get_range(quoted)\n        replacement =\n          {{:., dot_meta, [{:__aliases__, alias_meta, [:String]}, :to_existing_atom]}, call_meta, args}\n          |> Sourceror.to_string()\n\n        patch = %{range: range, change: replacement}\n        {quoted, [patch | patches]}\n\n      quoted, patches ->\n        {quoted, patches}\n    end)\n\n  assert Sourceror.patch_string(source, patches) ==\n    \"\"\"\n    case foo do\n      nil ->         :bar\n      _ ->\n\n          String.to_existing_atom(foo)\n\n          end\\\n    \"\"\"\nend\n```\n\nYou have to keep in mind that:\n\n1. If you patch a node that has inner code, like replacing a full `case`, then\n   the contents of the node will be reformatted as well.\n\n2. At the moment, Sourceror won't check for conflicts in the patches ranges, so\n   care needs to be taken to not produce conflicting patches. You may need to do\n   a number of `parse -> patch -> reparse` if you find yourself generating\n   conflicting patches.\n\nSome of the most common patching operations are available in the\n[Sourceror.Patch](https://hexdocs.pm/sourceror/Sourceror.Patch.html) module","ref":"Sourceror.html#module-patching-the-source-code"},{"type":"module","title":"Background - Sourceror","doc":"There have been several attempts at source code manipulation in the Elixir\ncommunity. Thanks to its metaprogramming features, Elixir provides builtin tools\nthat let us get the AST of any Elixir code, but when it comes to turning the AST\nback to code as text, we had limited options. `Macro.to_string/2` is a thing,\nbut the produced code is generally ugly, mostly because of the extra parenthesis\nor because it turns string interpolations into calls to erlang modules, to name\nsome examples. This meant that, even if we could use `Macro.to_string/2` to get\na string and then give that to the Elixir formatter `Code.format_string!/2`, the\noutput would still be suboptimal, as the formatter is not designed to change the\nsemantics of the code, only to pretty print it. For example, call to erlang\nmodules would be kept as is instead of being turned back to interpolations.\n\nWe also had the additional problem of comments being discarded by the tokenizer,\nand literals not having information like line numbers or delimiter characters.\nThis makes the regular AST too lossy to be useful if what we want is to\nmanipulate the source code, because we need as much information as possible to\nbe able to stay as close to the source as possible. There have been several\nproposal in the past to bring all this information to the Elixir AST, but they\nall meant a change that would either break macros due to the addition of new\ntypes of AST nodes, or making a compromise in core Elixir itself by storing\ncomments in the nods metadata. [This\ndiscussion](https://groups.google.com/u/1/g/elixir-lang-core/c/GM0yM5Su1Zc/m/poIKsiEVDQAJ)\nin the Elixir mailing list highlights the various issues faced when deciding if\nand how the comments would be preserved. Arjan Scherpenisse also did a\n[talk](https://www.youtube.com/watch?v=aM0BLWgr0g4&t=117s) where he discusses\nabout the problems of using the standard Elixir AST to build refactoring tools.\n\nDespite of all these issues, the Elixir formatter is still capable of\nmanipulating the source code to pretty print it. Under the hood it does some\nneat tricks to have all this information available: on one hand, it tells the\ntokenizer to extract the comments from the source code and keep it at hand(not\nin the AST itself, but as a separate data structure), and on the other hand it\ntells the parser to wrap literals in block nodes so metadata can be preserved.\nOnce it has all it needs, it can start converting the AST and comments into an\nalgebra document, and ultimately convert that to a string. This functionality\nwas private, and if we wanted to do it ourselves we would have to replicate or\nvendor the Elixir formatter with all its more than 2000 lines of code. This\napproach was explored by Wojtek Mach in\n[wojtekmach/fix](https://github.com/wojtekmach/fix), but it involved vendoring\nthe elixir Formatter code, was tightly coupled to the formatting process, and\nany change in Elixir would break the code.\n\nSince Elixir 1.13 this functionality from the formatter was finally exposed via\nthe `Code.string_to_quoted_with_comments/2` and `Code.quoted_to_algebra/2`\nfunctions. The former gives us access to the list of comments in a shape the\nElixir formatter is able to use, and the latter lets us turn _any arbitrary\nElixir AST_ into an algebra document. If we also give it the list of comments,\nit will merge them together, allowing us to format AST _and_ preserve the\ncomments. Now all we need to care about is of manipulating the AST, and let the\nformatter do the rest.","ref":"Sourceror.html#module-background"},{"type":"function","title":"Sourceror.append_comments/3","doc":"Appends comments to the leading or trailing comments of a node.","ref":"Sourceror.html#append_comments/3"},{"type":"function","title":"Sourceror.compare_positions/2","doc":"Compares two positions.\n\nReturns `:gt` if the first position comes after the second one, and `:lt` for\nvice versa. If the two positions are equal, `:eq` is returned.\n\n`nil` values for lines or columns are coalesced to `0` for integer\ncomparisons.","ref":"Sourceror.html#compare_positions/2"},{"type":"function","title":"Sourceror.correct_lines/3","doc":"Shifts the line numbers of the node or metadata by the given `line_correction`.\n\nThis function will update the `:line`, `:closing`, `:do`, `:end` and\n`:end_of_expression` line numbers of the node metadata if such fields are\npresent.","ref":"Sourceror.html#correct_lines/3"},{"type":"function","title":"Sourceror.get_args/1","doc":"Returns the arguments of the node.\n\n    iex> Sourceror.get_args({:foo, [], [{:__block__, [], [:ok]}]})\n    [{:__block__, [], [:ok]}]","ref":"Sourceror.html#get_args/1"},{"type":"function","title":"Sourceror.get_column/2","doc":"Returns the column of a node. If none is found, the default value is\nreturned(defaults to 1).\n\nA default of `nil` may also be provided if the column number is meant to be\ncoalesced with a value that is not known upfront.\n\n    iex> Sourceror.get_column({:foo, [column: 5], []})\n    5\n\n    iex> Sourceror.get_column({:foo, [], []}, 3)\n    3","ref":"Sourceror.html#get_column/2"},{"type":"function","title":"Sourceror.get_end_line/2","doc":"Returns the line where the given node ends. It recursively checks for `end`,\n`closing` and `end_of_expression` line numbers. If none is found, the default\nvalue is returned(defaults to 1).\n\n    iex> Sourceror.get_end_line({:foo, [end: [line: 4]], []})\n    4\n\n    iex> Sourceror.get_end_line({:foo, [closing: [line: 2]], []})\n    2\n\n    iex> Sourceror.get_end_line({:foo, [end_of_expression: [line: 5]], []})\n    5\n\n    iex> Sourceror.get_end_line({:foo, [closing: [line: 2], end: [line: 4]], []})\n    4\n\n    iex> \"\"\"\n    ...> alias Foo.{\n    ...>   Bar\n    ...> }\n    ...> \"\"\" |> Sourceror.parse_string!() |> Sourceror.get_end_line()\n    3","ref":"Sourceror.html#get_end_line/2"},{"type":"function","title":"Sourceror.get_end_position/2","doc":"Returns the end position of the quoted expression. It recursively checks for\n`end`, `closing` and `end_of_expression` positions. If none is found, the\ndefault value is returned(defaults to `[line: 1, column: 1]`).\n\nNote that this position is the literal contents of the metadata found\nin the node. For accurate end positions, use `get_range/1`.\n\n    iex> quoted = ~S\"\"\"\n    ...> A.{\n    ...>   B\n    ...> }\n    ...> \"\"\" |>  Sourceror.parse_string!()\n    iex> Sourceror.get_end_position(quoted)\n    [line: 3, column: 1]\n\n    iex> quoted = ~S\"\"\"\n    ...> foo do\n    ...>   :ok\n    ...> end\n    ...> \"\"\" |>  Sourceror.parse_string!()\n    iex> Sourceror.get_end_position(quoted)\n    [line: 3, column: 1]\n\n    iex> quoted = ~S\"\"\"\n    ...> foo(\n    ...>   :a,\n    ...>   :b\n    ...>    )\n    ...> \"\"\" |>  Sourceror.parse_string!()\n    iex> Sourceror.get_end_position(quoted)\n    [line: 4, column: 4]","ref":"Sourceror.html#get_end_position/2"},{"type":"function","title":"Sourceror.get_line/2","doc":"Returns the line of a node. If none is found, the default value is\nreturned(defaults to 1).\n\nA default of `nil` may also be provided if the line number is meant to be\ncoalesced with a value that is not known upfront.\n\n    iex> Sourceror.get_line({:foo, [line: 5], []})\n    5\n\n    iex> Sourceror.get_line({:foo, [], []}, 3)\n    3","ref":"Sourceror.html#get_line/2"},{"type":"function","title":"Sourceror.get_meta/1","doc":"Returns the metadata of the given node.\n\n    iex> Sourceror.get_meta({:foo, [line: 5], []})\n    [line: 5]","ref":"Sourceror.html#get_meta/1"},{"type":"function","title":"Sourceror.get_range/2","doc":"Gets the range used by the given quoted expression in the source code.\n\nThe quoted expression must have at least line and column metadata, otherwise\nit is not possible to calculate an accurate range, or to calculate it at all.\nAdditionally, certain syntax constructs desugar into ASTs without a\nmeaningful range. In these cases, `get_range/1` returns `nil`.\n\nThis function is most useful when used after `Sourceror.parse_string/1`,\nbefore any kind of modification to the AST.\n\n    iex> quoted = ~S\"\"\"\n    ...> def foo do\n    ...>   :ok\n    ...> end\n    ...> \"\"\" |> Sourceror.parse_string!()\n    iex> Sourceror.get_range(quoted)\n    %Sourceror.Range{start: [line: 1, column: 1], end: [line: 3, column: 4]}\n\n    iex> quoted = ~S\"\"\"\n    ...> Foo.{\n    ...>   Bar\n    ...> }\n    ...> \"\"\" |> Sourceror.parse_string!()\n    iex> Sourceror.get_range(quoted)\n    %Sourceror.Range{start: [line: 1, column: 1], end: [line: 3, column: 2]}","ref":"Sourceror.html#get_range/2"},{"type":"function","title":"Options - Sourceror.get_range/2","doc":"- `:include_comments` - When `true`, it includes the comments into the range. Defaults to `false`.\n\n```elixir\niex> ~S\"\"\"\n...> # Foo\n...> :baz # Bar\n...> \"\"\"\n...> |> Sourceror.parse_string!()\n...> |> Sourceror.get_range(include_comments: true)\n%Sourceror.Range{start: [line: 1, column: 1], end: [line: 2, column: 11]}\n```","ref":"Sourceror.html#get_range/2-options"},{"type":"function","title":"Sourceror.get_start_position/2","doc":"Returns the start metadata position of a node.\n\nNote that this position is the literal contents of the metadata found\nin the node. For accurate start positions, use `get_range/1`.\n\n    iex> quoted = Sourceror.parse_string!(\" :foo\")\n    iex> Sourceror.get_start_position(quoted)\n    [line: 1, column: 2]\n\n    iex> quoted = Sourceror.parse_string!(\"\\n\\nfoo()\")\n    iex> Sourceror.get_start_position(quoted)\n    [line: 3, column: 1]\n\n    iex> quoted = Sourceror.parse_string!(\"Foo.{Bar}\")\n    iex> Sourceror.get_start_position(quoted)\n    [line: 1, column: 1]\n\n    iex> quoted = Sourceror.parse_string!(\"foo[:bar]\")\n    iex> Sourceror.get_start_position(quoted)\n    [line: 1, column: 1]\n\n    iex> quoted = Sourceror.parse_string!(\"foo(:bar)\")\n    iex> Sourceror.get_start_position(quoted)\n    [line: 1, column: 1]","ref":"Sourceror.html#get_start_position/2"},{"type":"function","title":"Sourceror.parse_expression/2","doc":"Parses a single expression from the given string. It tries to parse on a\nper-line basis.\n\nReturns `{:ok, quoted, rest}` on success or `{:error, source}` on error.","ref":"Sourceror.html#parse_expression/2"},{"type":"function","title":"Examples - Sourceror.parse_expression/2","doc":"iex> ~S\"\"\"\n    ...> 42\n    ...>\n    ...> :ok\n    ...> \"\"\" |> Sourceror.parse_expression()\n    {:ok, {:__block__, [trailing_comments: [], leading_comments: [],\n                        token: \"42\", line: 2, column: 1], [42]}, \"\\n:ok\"}","ref":"Sourceror.html#parse_expression/2-examples"},{"type":"function","title":"Options - Sourceror.parse_expression/2","doc":"* `:from_line` - The line at where the parsing should start. Defaults to `1`.","ref":"Sourceror.html#parse_expression/2-options"},{"type":"function","title":"Sourceror.parse_string/1","doc":"Parses the source code into an extended AST suitable for source manipulation\nas described in `Code.quoted_to_algebra/2`.\n\nTwo additional fields are added to nodes metadata:\n  * `:leading_comments` - a list holding the comments found *before* the node.\n  * `:trailing_comments` - a list holding the comments found before the end of\n    the node. For example, comments right before the `end` keyword.\n\nComments are the same maps returned by `Code.string_to_quoted_with_comments/2`.","ref":"Sourceror.html#parse_string/1"},{"type":"function","title":"Sourceror.parse_string!/1","doc":"Same as `parse_string/1` but raises on error.","ref":"Sourceror.html#parse_string!/1"},{"type":"function","title":"Sourceror.patch_string/2","doc":"Applies one or more patches to the given string.\n\nA patch is a map containing the following values:\n\n  * `:range` - a map containing `:start` and `:end` keys, whose values are\n    keyword lists containing the `:line` and `:column` representing the\n    boundary of the patch.\n  * `:change` - the string being patched in or a function that takes the\n    text of the patch range and returns the replacement.\n  * `:preserve_indentation` (default `true`) - whether to automatically\n    correct the indentation of the patch string to preserve the indentation\n    level of the patch range (see examples below)\n\nNote that `:line` and `:column` start at 1 and represent a cursor positioned\nbefore the targeted position. For instance, here's how you would select the\nstring `\"ToBePatched\"` in the following example:\n\n    defmodule ToBePatched do\n    #         ^          ^\n    # col     11         22\n\n    %{range: %{start: [line: 1, column: 11], end: [line: 1, column: 22]}}\n\nRanges are usually derived from parsed AST nodes. See `get_range/2` for more.\n\nPatches are applied bottom-up, such that patches to the beginning of the\nstring do not interfere with the line/column of patches that come later.\nHowever, no checks are done for overlapping ranges, so take care to pass in\nnon-overlapping patches.","ref":"Sourceror.html#patch_string/2"},{"type":"function","title":"Examples - Sourceror.patch_string/2","doc":"iex> original = ~S\"\"\"\n    ...> if not allowed? do\n    ...>   raise \"Not allowed!\"\n    ...> end\n    ...> \"\"\"\n    iex> patch = %Sourceror.Patch{\n    ...>   change: \"unless allowed? do\\n  raise \\\"Not allowed!\\\"\\nend\",\n    ...>   range: %Sourceror.Range{start: [line: 1, column: 1], end: [line: 3, column: 4]}\n    ...> }\n    iex> Sourceror.patch_string(original, [patch])\n    ~S\"\"\"\n    unless allowed? do\n      raise \"Not allowed!\"\n    end\n    \"\"\"\n\nA range can also be a function, in which case the original text in the patch\nrange will be given as an argument:\n\n    iex> original = ~S\"\"\"\n    ...> hello :world\n    ...> \"\"\"\n    iex> patch = %Sourceror.Patch{\n    ...>   change: &String.upcase/1,\n    ...>   range: %Sourceror.Range{start: [line: 1, column: 7], end: [line: 1, column: 13]}\n    ...> }\n    iex> Sourceror.patch_string(original, [patch])\n    ~S\"\"\"\n    hello :WORLD\n    \"\"\"\n\nBy default, lines after the first line of the patch will be indented relative to\nthe indentation level at the start of the range:\n\n    iex> original = ~S\"\"\"\n    ...> outer do\n    ...>   inner(foo do\n    ...>     :original\n    ...>   end)\n    ...> end\n    ...> \"\"\"\n    iex> patch = %Sourceror.Patch{\n    ...>   change: \"bar do\\n  :replacement\\nend\",\n    ...>   range: %Sourceror.Range{start: [line: 2, column: 9], end: [line: 4, column: 6]}\n    ...> }\n    iex> Sourceror.patch_string(original, [patch])\n    ~S\"\"\"\n    outer do\n      inner(bar do\n        :replacement\n      end)\n    end\n    \"\"\"\n\nIf you don't want this behavior, you can add `:preserve_indentation: false` to\nyour patch:\n\n    iex> original = ~S\"\"\"\n    ...> outer do\n    ...>   inner(foo do\n    ...>     :original\n    ...>   end)\n    ...> end\n    ...> \"\"\"\n    iex> patch = %Sourceror.Patch{\n    ...>   change: \"bar do\\n  :replacement\\nend\",\n    ...>   range: %Sourceror.Range{start: [line: 2, column: 9], end: [line: 4, column: 6]},\n    ...>   preserve_indentation: false\n    ...> }\n    iex> Sourceror.patch_string(original, [patch])\n    ~S\"\"\"\n    outer do\n      inner(bar do\n      :replacement\n    end)\n    end\n    \"\"\"","ref":"Sourceror.html#patch_string/2-examples"},{"type":"function","title":"Sourceror.postwalk/2","doc":"Performs a depth-first post-order traversal of a quoted expression.\n\nSee `postwalk/3` for more information.","ref":"Sourceror.html#postwalk/2"},{"type":"function","title":"Sourceror.postwalk/3","doc":"Performs a depth-first post-order traversal of a quoted expression with an\naccumulator.\n\n`fun` is a function that will receive the current node as a first argument and\nthe traversal state as the second one. It must return a `{quoted, state}`,\nin the same way it would return `{quoted, acc}` when using `Macro.postwalk/3`.\n\nThe state is a map with the following keys:\n  * `:acc` - The accumulator. Defaults to `nil` if none is given.","ref":"Sourceror.html#postwalk/3"},{"type":"function","title":"Sourceror.prepend_comments/3","doc":"Prepends comments to the leading or trailing comments of a node.","ref":"Sourceror.html#prepend_comments/3"},{"type":"function","title":"Sourceror.prewalk/2","doc":"Performs a depth-first pre-order traversal of a quoted expression.\n\nSee `prewalk/3` for more information.","ref":"Sourceror.html#prewalk/2"},{"type":"function","title":"Sourceror.prewalk/3","doc":"Performs a depth-first pre-order traversal of a quoted expression with an\naccumulator.\n\n`fun` is a function that will receive the current node as a first argument and\nthe traversal state as the second one. It must return a `{quoted, state}`,\nin the same way it would return `{quoted, acc}` when using `Macro.prewalk/3`.\n\nThe state is a map with the following keys:\n  * `:acc` - The accumulator. Defaults to `nil` if none is given.","ref":"Sourceror.html#prewalk/3"},{"type":"macro","title":"Sourceror.quoted_to_algebra/2","doc":"A wrapper around `Code.quoted_to_algebra/2` for compatibility with pre 1.13\nElixir versions.","ref":"Sourceror.html#quoted_to_algebra/2"},{"type":"macro","title":"Sourceror.string_to_quoted/2","doc":"A wrapper around `Code.string_to_quoted_with_comments/2` for compatibility\nwith pre 1.13 Elixir versions.","ref":"Sourceror.html#string_to_quoted/2"},{"type":"macro","title":"Sourceror.string_to_quoted!/2","doc":"A wrapper around `Code.string_to_quoted_with_comments!/2` for compatibility\nwith pre 1.13 Elixir versions.","ref":"Sourceror.html#string_to_quoted!/2"},{"type":"function","title":"Sourceror.to_string/2","doc":"Converts a quoted expression to a string.\n\nThe comments line number will be ignored and the line number of the associated\nnode will be used when formatting the code.\n\nIf you're using Elixir >=1.13, the `locals_without_parens` from your project\nformatter configuration will be used. You can pass a different set of options\nby using the `locals_without_parens` option. If you want to disable that option\nentirely, use `locals_without_parens: []`.","ref":"Sourceror.html#to_string/2"},{"type":"function","title":"Options - Sourceror.to_string/2","doc":"* `:indent` - how many indentations to insert at the start of each line.\n    Note that this only prepends the indents without checking the indentation\n    of nested blocks. Defaults to `0`.\n\n  * `:indent_type` - the type of indentation to use. It can be one of `:spaces`,\n    `:single_space` or `:tabs`. Defaults to `:spaces`.\n\n  * `:format` - if set to `:splicing`, if the quoted expression is a list, it\n    will strip the square brackets. This is useful to print a single element\n    of a keyword list.\n\n  * `:quoted_to_algebra` - expects a function of the type\n    `(Macro.t(), keyword -> Inspect.Algebra.t())` to convert the given quoted\n    expression to an algebra document.\n\nFor more options see `Code.format_string!/1` and `Code.quoted_to_algebra/2`.","ref":"Sourceror.html#to_string/2-options"},{"type":"function","title":"Sourceror.update_args/2","doc":"Updates the arguments for the given node.\n\n    iex> node = {:foo, [line: 1], [{:__block__, [line: 1], [2]}]}\n    iex> updater = fn args -> Enum.map(args, &Sourceror.correct_lines(&1, 2)) end\n    iex> Sourceror.update_args(node, updater)\n    {:foo, [line: 1], [{:__block__, [line: 3], [2]}]}","ref":"Sourceror.html#update_args/2"},{"type":"type","title":"Sourceror.comment/0","doc":"","ref":"Sourceror.html#t:comment/0"},{"type":"type","title":"Sourceror.position/0","doc":"","ref":"Sourceror.html#t:position/0"},{"type":"type","title":"Sourceror.traversal_function/0","doc":"","ref":"Sourceror.html#t:traversal_function/0"},{"type":"module","title":"Sourceror.Comments","doc":"Utilities to merge an un-merge comments and quoted expressions.","ref":"Sourceror.Comments.html"},{"type":"function","title":"Sourceror.Comments.extract_comments/2","doc":"Does the opposite of `merge_comments/2`, it extracts the comments from the\nquoted expression and returns both as a `{quoted, comments}` tuple.","ref":"Sourceror.Comments.html#extract_comments/2"},{"type":"function","title":"Sourceror.Comments.merge_comments/2","doc":"Merges the comments into the given quoted expression.\n\nThe comments are inserted into the metadata of their closest node. Comments in\nthe same line of before a node are inserted into the `:leading_comments` field\nwhile comments that are right before an `end` keyword are inserted into the\n`:trailing_comments` field.","ref":"Sourceror.Comments.html#merge_comments/2"},{"type":"module","title":"Sourceror.Identifier","doc":"Functions to identify an classify forms and quoted expressions.","ref":"Sourceror.Identifier.html"},{"type":"macro","title":"Sourceror.Identifier.is_atomic_literal/1","doc":"Checks if the given quoted form is an atomic literal in the AST.\n\nThis set includes numbers, atoms, and strings, but not collections like\ntuples, lists, or maps.\n\nThis guard returns `true` for literals that are the only elements inside\nof a `:__block__`, such as `{:__block__, [], [:literal]}`.","ref":"Sourceror.Identifier.html#is_atomic_literal/1"},{"type":"macro","title":"Examples - Sourceror.Identifier.is_atomic_literal/1","doc":"iex> is_atomic_literal(1)\n    true\n\n    iex> is_atomic_literal(1.0)\n    true\n\n    iex> is_atomic_literal(:foo)\n    true\n\n    iex> is_atomic_literal(\"foo\")\n    true\n\n    iex> is_atomic_literal({:__block__, [], [1]})\n    true\n\n    iex> is_atomic_literal({:__block__, [], [1, 2]})\n    false\n\n    iex> is_atomic_literal({:__block__, [], [{:node, [], nil}]})\n    false\n\n    iex> is_atomic_literal('foo')\n    false","ref":"Sourceror.Identifier.html#is_atomic_literal/1-examples"},{"type":"macro","title":"Sourceror.Identifier.is_binary_op/1","doc":"Checks if the given identifier is a binary op.","ref":"Sourceror.Identifier.html#is_binary_op/1"},{"type":"macro","title":"Examples - Sourceror.Identifier.is_binary_op/1","doc":"iex> is_binary_op(:+)\n    true","ref":"Sourceror.Identifier.html#is_binary_op/1-examples"},{"type":"macro","title":"Sourceror.Identifier.is_call/1","doc":"Checks if the given quoted form is a call.\n\nCalls are any form of the shape `{form, metadata, args}` where args is\na list, with the exception of blocks and aliases, which are identified\nby the forms `:__block__` and `:__aliases__`.","ref":"Sourceror.Identifier.html#is_call/1"},{"type":"macro","title":"Examples - Sourceror.Identifier.is_call/1","doc":"iex> \"node()\" |> Sourceror.parse_string!() |> is_call()\n    true\n\n    iex> \"Kernel.node()\" |> Sourceror.parse_string!() |> is_call()\n    true\n\n    iex> \"%{}\" |> Sourceror.parse_string!() |> is_call()\n    true\n\n    iex> \"@attr\" |> Sourceror.parse_string!() |> is_call()\n    true\n\n    iex> \"node\" |> Sourceror.parse_string!() |> is_call()\n    false\n\n    iex> \"1\" |> Sourceror.parse_string!() |> is_call()\n    false\n\n    iex> \"(1; 2)\" |> Sourceror.parse_string!() |> is_call()\n    false\n\n    iex> \"Macro.Env\" |> Sourceror.parse_string!() |> is_call()\n    false","ref":"Sourceror.Identifier.html#is_call/1-examples"},{"type":"macro","title":"Sourceror.Identifier.is_identifier/1","doc":"Checks if the given quoted form is an identifier, such as a variable.","ref":"Sourceror.Identifier.html#is_identifier/1"},{"type":"macro","title":"Examples - Sourceror.Identifier.is_identifier/1","doc":"iex> \"node\" |> Sourceror.parse_string!() |> is_identifier()\n    true\n\n    iex> \"node()\" |> Sourceror.parse_string!() |> is_identifier()\n    false\n\n    iex> \"1\" |> Sourceror.parse_string!() |> is_identifier()\n    false","ref":"Sourceror.Identifier.html#is_identifier/1-examples"},{"type":"macro","title":"Sourceror.Identifier.is_pipeline_op/1","doc":"Checks if the given identifier is a pipeline operator.","ref":"Sourceror.Identifier.html#is_pipeline_op/1"},{"type":"macro","title":"Examples - Sourceror.Identifier.is_pipeline_op/1","doc":"iex> is_pipeline_op(:|>)\n    true","ref":"Sourceror.Identifier.html#is_pipeline_op/1-examples"},{"type":"macro","title":"Sourceror.Identifier.is_qualified_call/1","doc":"Checks if the given quoted form is a qualified call.\n\nAll unqualified calls would also return `true` if passed to `is_call/1`,\nbut they have the shape `{{:., dot_metadata, dot_args}, metadata, args}`.","ref":"Sourceror.Identifier.html#is_qualified_call/1"},{"type":"macro","title":"Examples - Sourceror.Identifier.is_qualified_call/1","doc":"iex> \"Kernel.node()\" |> Sourceror.parse_string!() |> is_qualified_call()\n    true\n\n    iex> \"__MODULE__.node()\" |> Sourceror.parse_string!() |> is_qualified_call()\n    true\n\n    iex> \"foo.()\" |> Sourceror.parse_string!() |> is_qualified_call()\n    true\n\n    iex> \"foo.bar()\" |> Sourceror.parse_string!() |> is_qualified_call()\n    true\n\n    iex> \"node()\" |> Sourceror.parse_string!() |> is_qualified_call()\n    false\n\n    iex> \"%{}\" |> Sourceror.parse_string!() |> is_qualified_call()\n    false\n\n    iex> \"@attr\" |> Sourceror.parse_string!() |> is_qualified_call()\n    false\n\n    iex> \"1\" |> Sourceror.parse_string!() |> is_qualified_call()\n    false\n\n    iex> \"(1; 2)\" |> Sourceror.parse_string!() |> is_qualified_call()\n    false\n\n    iex> \"Macro.Env\" |> Sourceror.parse_string!() |> is_qualified_call()\n    false","ref":"Sourceror.Identifier.html#is_qualified_call/1-examples"},{"type":"macro","title":"Sourceror.Identifier.is_reserved_block_name/1","doc":"Checks if the given quoted form is a reserved block name.","ref":"Sourceror.Identifier.html#is_reserved_block_name/1"},{"type":"macro","title":"Examples - Sourceror.Identifier.is_reserved_block_name/1","doc":"iex> is_reserved_block_name(:do)\n    true\n\n    iex> is_reserved_block_name(:else)\n    true\n\n    iex> is_reserved_block_name(:catch)\n    true\n\n    iex> is_reserved_block_name(:after)\n    true\n\n    iex> is_reserved_block_name(:rescue)\n    true\n\n    iex> is_reserved_block_name(:foo)\n    false","ref":"Sourceror.Identifier.html#is_reserved_block_name/1-examples"},{"type":"macro","title":"Sourceror.Identifier.is_unary_op/1","doc":"Checks if the given identifier is an unary op.","ref":"Sourceror.Identifier.html#is_unary_op/1"},{"type":"macro","title":"Examples - Sourceror.Identifier.is_unary_op/1","doc":"iex> is_unary_op(:+)\n    true","ref":"Sourceror.Identifier.html#is_unary_op/1-examples"},{"type":"macro","title":"Sourceror.Identifier.is_unqualified_call/1","doc":"Checks if the given quoted form is an unqualified call.\n\nAll unqualified calls would also return `true` if passed to `is_call/1`,\nbut they have the shape `{atom, metadata, args}`.","ref":"Sourceror.Identifier.html#is_unqualified_call/1"},{"type":"macro","title":"Examples - Sourceror.Identifier.is_unqualified_call/1","doc":"iex> \"node()\" |> Sourceror.parse_string!() |> is_unqualified_call()\n    true\n\n    iex> \"%{}\" |> Sourceror.parse_string!() |> is_unqualified_call()\n    true\n\n    iex> \"@attr\" |> Sourceror.parse_string!() |> is_unqualified_call()\n    true\n\n    iex> \"node\" |> Sourceror.parse_string!() |> is_unqualified_call()\n    false\n\n    iex> \"1\" |> Sourceror.parse_string!() |> is_unqualified_call()\n    false\n\n    iex> \"(1; 2)\" |> Sourceror.parse_string!() |> is_unqualified_call()\n    false\n\n    iex> \"Macro.Env\" |> Sourceror.parse_string!() |> is_unqualified_call()\n    false","ref":"Sourceror.Identifier.html#is_unqualified_call/1-examples"},{"type":"function","title":"Sourceror.Identifier.valid_alias?/1","doc":"Checks if the given atom is a valid module alias.","ref":"Sourceror.Identifier.html#valid_alias?/1"},{"type":"function","title":"Examples - Sourceror.Identifier.valid_alias?/1","doc":"iex> valid_alias?(Foo)\n    true\n    iex> valid_alias?(:foo)\n    false","ref":"Sourceror.Identifier.html#valid_alias?/1-examples"},{"type":"module","title":"Sourceror.Patch","doc":"Functions that generate patches for common operations.\n\nFunctions in this module assume that the AST was parsed using Sourceror\nfunctions and that it wasn't modified. If you changed the tree before calling\n`Sourceror.Patch` functions, then the patch ranges are not guaranteed to match\n1:1 with the original source code.","ref":"Sourceror.Patch.html"},{"type":"function","title":"Sourceror.Patch.new/3","doc":"Creates a new patch.","ref":"Sourceror.Patch.html#new/3"},{"type":"function","title":"Sourceror.Patch.rename_call/2","doc":"Renames a qualified or unqualified function call.\n\n    iex> original = \"String.to_atom(foo)\"\n    iex> ast = Sourceror.parse_string!(original)\n    iex> patch = Sourceror.Patch.rename_call(ast, :to_existing_atom)\n    iex> Sourceror.patch_string(original, [patch])\n    \"String.to_existing_atom(foo)\"\n\nIf the call is a sigil, you only need to provide the replacement letter:\n\n    iex> original = \"~H(foo)\"\n    iex> ast = Sourceror.parse_string!(original)\n    iex> patch = Sourceror.Patch.rename_call(ast, :F)\n    iex> Sourceror.patch_string(original, [patch])\n    \"~F(foo)\"","ref":"Sourceror.Patch.html#rename_call/2"},{"type":"function","title":"Sourceror.Patch.rename_identifier/2","doc":"Renames an identifier(ie a variable name).","ref":"Sourceror.Patch.html#rename_identifier/2"},{"type":"function","title":"Examples - Sourceror.Patch.rename_identifier/2","doc":"iex> original = \"foo\"\n    iex> ast = Sourceror.parse_string!(original)\n    iex> patch = Sourceror.Patch.rename_identifier(ast, :bar)\n    iex> Sourceror.patch_string(original, [patch])\n    \"bar\"","ref":"Sourceror.Patch.html#rename_identifier/2-examples"},{"type":"function","title":"Sourceror.Patch.rename_kw_keys/2","doc":"Generates patches that rename the keys of a keyword list.\n\nThe replacements is a keyword list, with the keys to replace as keys, and the\nreplacement as the value.","ref":"Sourceror.Patch.html#rename_kw_keys/2"},{"type":"function","title":"Examples - Sourceror.Patch.rename_kw_keys/2","doc":"iex> original = \"[a: b, c: d, e: f]\"\n    iex> ast = Sourceror.parse_string!(original)\n    iex> patches = Sourceror.Patch.rename_kw_keys(ast, a: :foo, e: :bar)\n    iex> Sourceror.patch_string(original, patches)\n    \"[foo: b, c: d, bar: f]\"","ref":"Sourceror.Patch.html#rename_kw_keys/2-examples"},{"type":"function","title":"Sourceror.Patch.replace/2","doc":"Generates a patch that replaces the node with the given replacement.","ref":"Sourceror.Patch.html#replace/2"},{"type":"function","title":"Examples - Sourceror.Patch.replace/2","doc":"iex> original = \"foo\"\n    iex> ast = Sourceror.parse_string!(original)\n    iex> patch = Sourceror.Patch.replace(ast, \"bar\")\n    iex> Sourceror.patch_string(original, [patch])\n    \"bar\"","ref":"Sourceror.Patch.html#replace/2-examples"},{"type":"type","title":"Sourceror.Patch.change/0","doc":"","ref":"Sourceror.Patch.html#t:change/0"},{"type":"type","title":"Sourceror.Patch.t/0","doc":"","ref":"Sourceror.Patch.html#t:t/0"},{"type":"module","title":"Sourceror.Range","doc":"Represents a start/end position in a source file as `:line` and `:column`\nkeyword lists.","ref":"Sourceror.Range.html"},{"type":"function","title":"Sourceror.Range.new/2","doc":"Creates a new range struct.","ref":"Sourceror.Range.html#new/2"},{"type":"function","title":"Examples - Sourceror.Range.new/2","doc":"iex> Sourceror.Range.new([line: 1, column: 1], [line: 1, column: 2])\n    %Sourceror.Range{start: [line: 1, column: 1], end: [line: 1, column: 2]}","ref":"Sourceror.Range.html#new/2-examples"},{"type":"type","title":"Sourceror.Range.t/0","doc":"","ref":"Sourceror.Range.html#t:t/0"},{"type":"module","title":"Sourceror.TraversalState","doc":"The state struct for Sourceror traversal functions.","ref":"Sourceror.TraversalState.html"},{"type":"type","title":"Sourceror.TraversalState.t/0","doc":"","ref":"Sourceror.TraversalState.html#t:t/0"},{"type":"module","title":"Sourceror.Zipper","doc":"Tree-like data structure that provides enhanced navigation and modification\nof an Elixir AST.\n\nThis implementation is based on Gérard Huet [Functional pearl: the zipper](https://www.st.cs.uni-saarland.de/edu/seminare/2005/advanced-fp/docs/huet-zipper.pdf)\nand Clojure's `clojure.zip` API.\n\nA zipper is a data structure that represents a location in a tree from the\nperspective of the current node, also called the *focus*.\n\nIt is represented by a struct containing a `:node` and `:path`, which is a\nprivate field used to track the position of the `:node` with regards to\nthe entire tree. The `:path` is an implementation detail that should be\nconsidered private.\n\nFor more information and examples, see the following guides:\n\n  * [Zippers](zippers.html) (an introduction)\n  * [Expand multi-alias syntax](expand_multi_alias.html) (an example)","ref":"Sourceror.Zipper.html"},{"type":"function","title":"Sourceror.Zipper.append_child/2","doc":"Inserts the `child` as the rightmost `child` of the `node` at this `zipper`,\nwithout moving.","ref":"Sourceror.Zipper.html#append_child/2"},{"type":"function","title":"Sourceror.Zipper.branch?/1","doc":"","ref":"Sourceror.Zipper.html#branch?/1"},{"type":"function","title":"Sourceror.Zipper.children/1","doc":"Returns a list of children of the `node`. Returns `nil` if the node is a leaf.","ref":"Sourceror.Zipper.html#children/1"},{"type":"function","title":"Sourceror.Zipper.do_find/3","doc":"","ref":"Sourceror.Zipper.html#do_find/3"},{"type":"function","title":"Sourceror.Zipper.down/1","doc":"Returns the `zipper` of the leftmost child of the `node` at this `zipper`, or\n`nil` if there's no children.","ref":"Sourceror.Zipper.html#down/1"},{"type":"function","title":"Sourceror.Zipper.find/3","doc":"Returns a `zipper` to the `node` that satisfies the `predicate` function, or\n`nil` if none is found.\n\nThe optional second parameters specifies the `direction`, defaults to\n`:next`.","ref":"Sourceror.Zipper.html#find/3"},{"type":"function","title":"Sourceror.Zipper.insert_child/2","doc":"Inserts the `child` as the leftmost `child` of the `node` at this `zipper`,\nwithout moving.","ref":"Sourceror.Zipper.html#insert_child/2"},{"type":"function","title":"Sourceror.Zipper.insert_left/2","doc":"Inserts the `child` as the left sibling of the `node` at this `zipper`,\nwithout moving. Raises an `ArgumentError` when attempting to insert a sibling\nat the top level.","ref":"Sourceror.Zipper.html#insert_left/2"},{"type":"function","title":"Sourceror.Zipper.insert_right/2","doc":"Inserts the `child` as the right sibling of the `node` at this `zipper`,\nwithout moving. Raises an `ArgumentError` when attempting to insert a sibling\nat the top level.","ref":"Sourceror.Zipper.html#insert_right/2"},{"type":"function","title":"Sourceror.Zipper.left/1","doc":"Returns the `zipper` of the left sibling of the `node` at this `zipper`, or\n`nil`.","ref":"Sourceror.Zipper.html#left/1"},{"type":"function","title":"Sourceror.Zipper.leftmost/1","doc":"Returns the leftmost sibling of the `node` at this `zipper`, or itself.","ref":"Sourceror.Zipper.html#leftmost/1"},{"type":"function","title":"Sourceror.Zipper.make_node/2","doc":"Returns a new branch `node`, given an existing `node` and new `children`.","ref":"Sourceror.Zipper.html#make_node/2"},{"type":"function","title":"Sourceror.Zipper.next/1","doc":"Returns the following `zipper` in depth-first pre-order.","ref":"Sourceror.Zipper.html#next/1"},{"type":"function","title":"Sourceror.Zipper.node/1","doc":"Returns the `node` at the `zipper`.","ref":"Sourceror.Zipper.html#node/1"},{"type":"function","title":"Sourceror.Zipper.prev/1","doc":"Returns the previous `zipper` in depth-first pre-order. If it's already at\nthe end, it returns `nil`.","ref":"Sourceror.Zipper.html#prev/1"},{"type":"function","title":"Sourceror.Zipper.remove/1","doc":"Removes the `node` at the zipper, returning the `zipper` that would have\npreceded it in a depth-first walk. Raises an `ArgumentError` when attempting\nto remove the top level `node`.","ref":"Sourceror.Zipper.html#remove/1"},{"type":"function","title":"Sourceror.Zipper.replace/2","doc":"Replaces the current `node` in the `zipper` with a new `node`.","ref":"Sourceror.Zipper.html#replace/2"},{"type":"function","title":"Sourceror.Zipper.right/1","doc":"Returns the zipper of the right sibling of the `nod`e at this `zipper`, or\nnil.","ref":"Sourceror.Zipper.html#right/1"},{"type":"function","title":"Sourceror.Zipper.rightmost/1","doc":"Returns the rightmost sibling of the `node` at this `zipper`, or itself.","ref":"Sourceror.Zipper.html#rightmost/1"},{"type":"function","title":"Sourceror.Zipper.root/1","doc":"Walks the `zipper` to the top of the current subtree and returns the that `node`.","ref":"Sourceror.Zipper.html#root/1"},{"type":"function","title":"Sourceror.Zipper.skip/2","doc":"Returns the `zipper` of the right sibling of the `node` at this `zipper`, or\nthe next `zipper` when no right sibling is available.\n\nThis allows to skip subtrees while traversing the siblings of a node.\n\nThe optional second parameters specifies the `direction`, defaults to\n`:next`.\n\nIf no right/left sibling is available, this function returns the same value as\n`next/1`/`prev/1`.\n\nThe function `skip/1` behaves like the `:skip` in `traverse_while/2` and\n`traverse_while/3`.","ref":"Sourceror.Zipper.html#skip/2"},{"type":"function","title":"Sourceror.Zipper.subtree/1","doc":"Returns a new `zipper` that is a subtree of the currently focused `node`.","ref":"Sourceror.Zipper.html#subtree/1"},{"type":"function","title":"Sourceror.Zipper.top/1","doc":"Walks the `zipper` to the top of the current subtree and returns that `zipper`.","ref":"Sourceror.Zipper.html#top/1"},{"type":"function","title":"Sourceror.Zipper.topmost/1","doc":"Walks the `zipper` to the topmost node, breaking out of any subtrees and returns the top-most `zipper`.","ref":"Sourceror.Zipper.html#topmost/1"},{"type":"function","title":"Sourceror.Zipper.topmost_root/1","doc":"Walks the `zipper` to the topmost node, breaking out of any subtrees and returns the root `node`.","ref":"Sourceror.Zipper.html#topmost_root/1"},{"type":"function","title":"Sourceror.Zipper.traverse/2","doc":"Traverses the tree in depth-first pre-order calling the given `fun` for each\n`node`. When the traversal is finished, the zipper will be back where it\nbegan.\n\nIf the `zipper` is not at the top, just the subtree will be traversed.\n\nThe function must return a `zipper`.","ref":"Sourceror.Zipper.html#traverse/2"},{"type":"function","title":"Sourceror.Zipper.traverse/3","doc":"Traverses the tree in depth-first pre-order calling the given `fun` for each\n`node` with an accumulator. When the traversal is finished, the zipper\nwill be back where it began.\n\nIf the `zipper` is not at the top, just the subtree will be traversed.","ref":"Sourceror.Zipper.html#traverse/3"},{"type":"function","title":"Sourceror.Zipper.traverse_while/2","doc":"Traverses the tree in depth-first pre-order calling the given `fun` for each\n`node`.\n\nThe traversing will continue if `fun` returns `{:cont, zipper}`, skipped for\n`{:skip, zipper}` and halted for `{:halt, zipper}`. When the traversal is\nfinished, the `zipper` will be back where it began.\n\nIf the `zipper` is not at the top, just the subtree will be traversed.\n\nThe function must return a `zipper`.","ref":"Sourceror.Zipper.html#traverse_while/2"},{"type":"function","title":"Sourceror.Zipper.traverse_while/3","doc":"Traverses the tree in depth-first pre-order calling the given `fun` for each\n`node` with an accumulator. When the traversal is finished, the `zipper`\nwill be back where it began.\n\nThe traversing will continue if `fun` returns `{:cont, zipper, acc}`, skipped\nfor `{:skip, zipper, acc}` and halted for `{:halt, zipper, acc}`\n\nIf the `zipper` is not at the top, just the subtree will be traversed.","ref":"Sourceror.Zipper.html#traverse_while/3"},{"type":"function","title":"Sourceror.Zipper.up/1","doc":"Returns the `zipper` for the parent `node` of the given `zipper`, or `nil` if\nthe `zipper` points to the root.","ref":"Sourceror.Zipper.html#up/1"},{"type":"function","title":"Sourceror.Zipper.update/2","doc":"Replaces the current `node` in the zipper with the result of applying `fun` to\nthe `node`.","ref":"Sourceror.Zipper.html#update/2"},{"type":"function","title":"Sourceror.Zipper.within/2","doc":"Runs the function `fun` on the subtree of the currently focused `node` and\nreturns the updated `zipper`.\n\n`fun` must return a zipper, which may be positioned at the top of the subtree.","ref":"Sourceror.Zipper.html#within/2"},{"type":"function","title":"Sourceror.Zipper.zip/1","doc":"Creates a `zipper` from a tree `node`.","ref":"Sourceror.Zipper.html#zip/1"},{"type":"opaque","title":"Sourceror.Zipper.path/0","doc":"","ref":"Sourceror.Zipper.html#t:path/0"},{"type":"type","title":"Sourceror.Zipper.t/0","doc":"","ref":"Sourceror.Zipper.html#t:t/0"},{"type":"type","title":"Sourceror.Zipper.tree/0","doc":"","ref":"Sourceror.Zipper.html#t:tree/0"},{"type":"module","title":"Sourceror.Zipper.Inspect","doc":"Provides `Sourceror.Zipper`'s implementation for the `Inspect` protocol.\n\nWhen inspecting a zipper, the default representation shows the current\nnode and provides indicators of the surrounding context without displaying\nthe full path, which can be very verbose for large ASTs.\n\n    iex> alias Sourceror.Zipper, as: Z\n    Sourceror.Zipper\n\n    iex> code = \"\"\"\n    ...> def my_function do\n    ...>   :ok\n    ...> end\\\n    ...> \"\"\"\n\n    iex> zipper = code |> Code.string_to_quoted!() |> Z.zip()\n    #Sourceror.Zipper<\n      #root\n      {:def, [line: 1], [{:my_function, [line: 1], nil}, [do: :ok]]}\n    >\n\n    iex> zipper |> Z.next()\n    #Sourceror.Zipper<\n      {:my_function, [line: 1], nil}\n      #...\n    >\n\nThis representation can be changed using `default_inspect_as/1` to set a\nglobal default or by using `:custom_options` with the `:zippers` key when\ninspecting. (See `Inspect.Opts` for more about inspect options.)\n\nZippers can be inspected in these formats:\n\n  * `:as_ast` (default, seen above) - display the current node as an AST\n  * `:as_code` - display the current node formatted as code\n  * `:raw` - display the raw `%Sourceror.Zipper{}` struct including the\n    `:path`.\n\nUsing the zipper defined above as an example:\n\n    iex> zipper |> inspect(custom_options: [zipper: :as_code]) |> IO.puts()\n    #Sourceror.Zipper<\n      #root\n      def my_function do\n        :ok\n      end\n    >\n\n    iex> zipper |> Z.next() |> inspect(custom_options: [zipper: :as_code]) |> IO.puts()\n    #Sourceror.Zipper \n\n    iex> zipper |> Z.next() |> inspect(custom_options: [zipper: :raw], pretty: true) |> IO.puts()\n    %Sourceror.Zipper{\n      node: {:my_function, [line: 1], nil},\n      path: %{\n        parent: %Sourceror.Zipper{\n          node: {:def, [line: 1], [{:my_function, [line: 1], nil}, [do: :ok]]},\n          path: nil\n        },\n        left: nil,\n        right: [[do: :ok]]\n      }\n    }","ref":"Sourceror.Zipper.Inspect.html"},{"type":"function","title":"Sourceror.Zipper.Inspect.default_inspect_as/1","doc":"Sets the default inspection format for zippers.","ref":"Sourceror.Zipper.Inspect.html#default_inspect_as/1"},{"type":"function","title":"Examples - Sourceror.Zipper.Inspect.default_inspect_as/1","doc":"Consider the following zipper:\n\n    iex> zipper\n    #Sourceror.Zipper<\n      #root\n      {:def, [line: 1], [{:my_function, [line: 1], nil}, [do: :ok]]}\n    >\n\n    iex> Sourceror.Zipper.Inspect.default_inspect_as(:as_code)\n    :ok\n    iex> zipper\n    #Sourceror.Zipper<\n      #root\n      def my_function do\n        :ok\n      end\n    >\n\n    iex> Sourceror.Zipper.Inspect.default_inspect_as(:raw)\n    :ok\n    iex> zipper\n    %Sourceror.Zipper{\n      node: {:def, [line: 1], [{:my_function, [line: 1], nil}, [do: :ok]]},\n      path: nil\n    }","ref":"Sourceror.Zipper.Inspect.html#default_inspect_as/1-examples"},{"type":"type","title":"Sourceror.Zipper.Inspect.inspect_as/0","doc":"Inspection formats for zippers.","ref":"Sourceror.Zipper.Inspect.html#t:inspect_as/0"},{"type":"extras","title":"Changelog","doc":"# Changelog for Sourceror v1.0\n\nAll notable changes to this project will be documented in this file.\n\nThe format is based on [Keep a Changelog](https://keepachangelog.com/en/1.0.0/),\nand this project adheres to [Semantic Versioning](https://semver.org/spec/v2.0.0.html).","ref":"changelog.html"},{"type":"extras","title":"v1.3.0 (2024-06-14) - Changelog","doc":"#","ref":"changelog.html#v1-3-0-2024-06-14"},{"type":"extras","title":"1. Enhancements - Changelog","doc":"- [Sourceror] Converted patch and range maps into `Patch` and `Range` structs\n- [Sourceror.Zipper] Added `within/2`\n  last elemetn in a block\n- [Sourceror.Zipper] Add `topmost` and `topmost_root`\n\n#","ref":"changelog.html#1-enhancements"},{"type":"extras","title":"2. Bug fixes - Changelog","doc":"- [Sourceror] Fix range calculations for heredocs\n- [Sourceror.Zipper] Fixed `remove/1` producing invalid AST when removing the\n- [Sourceror.Zipper] Fixed match on empty children list in `down/1`","ref":"changelog.html#2-bug-fixes"},{"type":"extras","title":"v1.2.1 (2024-05-23) - Changelog","doc":"#","ref":"changelog.html#v1-2-1-2024-05-23"},{"type":"extras","title":"1. Bug fixes - Changelog","doc":"- [Sourceror] Fixed line/column metadata for map literals.","ref":"changelog.html#1-bug-fixes"},{"type":"extras","title":"v1.2.0 (2024-05-22) - Changelog","doc":"#","ref":"changelog.html#v1-2-0-2024-05-22"},{"type":"extras","title":"1. Enhancements - Changelog","doc":"- [Sourceror.Zipper] Added `subtree/1` to get a zipper for the current node.","ref":"changelog.html#1-enhancements"},{"type":"extras","title":"v1.1.0 (2024-05-09) - Changelog","doc":"#","ref":"changelog.html#v1-1-0-2024-05-09"},{"type":"extras","title":"1. Bug fixes - Changelog","doc":"- [Sourceror] Fix trailing comments being misplaced","ref":"changelog.html#1-bug-fixes"},{"type":"extras","title":"v1.0.3 (2024-04-12) - Changelog","doc":"#","ref":"changelog.html#v1-0-3-2024-04-12"},{"type":"extras","title":"1. Bug fixes - Changelog","doc":"- [Sourceror] Fix `Sourceror.get_range` for interpolation nodes inside\n  strings and charlists","ref":"changelog.html#1-bug-fixes"},{"type":"extras","title":"v1.0.2 (2024-03-13) - Changelog","doc":"#","ref":"changelog.html#v1-0-2-2024-03-13"},{"type":"extras","title":"1. Bug fixes - Changelog","doc":"- [Sourceror] Fix `Sourceror.get_range` for binaries\n- [Sourceror] Fix `Sourceror.get_range` for `do`/`end` blocks that have\n  `end_of_expression` metadata","ref":"changelog.html#1-bug-fixes"},{"type":"extras","title":"v1.0.1 (2024-01-04) - Changelog","doc":"#","ref":"changelog.html#v1-0-1-2024-01-04"},{"type":"extras","title":"1. Bug fixes - Changelog","doc":"- [Sourceror] Fix `Sourceror.get_range` returning incorrect ranges for\n  anonymous functions with empty bodies","ref":"changelog.html#1-bug-fixes"},{"type":"extras","title":"v1.0.0 (2023-12-28) - Changelog","doc":"#","ref":"changelog.html#v1-0-0-2023-12-28"},{"type":"extras","title":"1. Bug fixes - Changelog","doc":"- [Sourceror] Add support for Elixir 1.16","ref":"changelog.html#1-bug-fixes"},{"type":"extras","title":"v0.13.0 (2023-08-23) - Changelog","doc":"The changelog for v0.14 releases can be found [in the 0.14\nbranch](https://github.com/doorgan/sourceror/blob/v0.14/CHANGELOG.md).","ref":"changelog.html#v0-13-0-2023-08-23"},{"type":"extras","title":"Contributing","doc":"# Contributing to Sourceror","ref":"contributing.html"},{"type":"extras","title":"Issues - Contributing","doc":"For issues like bug reports please file an issue in the [Issue\ntracker](https://github.com/doorgan/sourceror/issues).","ref":"contributing.html#issues"},{"type":"extras","title":"Proposals - Contributing","doc":"To propose new features to the library, please open a thread in the Discussions\nsection under the\n[Proposals](https://github.com/doorgan/sourceror/discussions/categories/proposals)\ncategory.\n\n# Development Contributions\n\nTake a look at the open issues\n[here](https://github.com/doorgan/sourceror/issues) and post that you plan on\nworking on an issue, so we can discuss approaches and make sure there isn't\nduplicated work happening.\n\nBefore submitting a pull request, make sure all of the following pass:\n\n```sh\nmix deps.get\nmix check\n```","ref":"contributing.html#proposals"},{"type":"extras","title":"License","doc":"Apache License\n                           Version 2.0, January 2004\n                        http://www.apache.org/licenses/\n\n   TERMS AND CONDITIONS FOR USE, REPRODUCTION, AND DISTRIBUTION\n\n   1. Definitions.\n\n      \"License\" shall mean the terms and conditions for use, reproduction,\n      and distribution as defined by Sections 1 through 9 of this document.\n\n      \"Licensor\" shall mean the copyright owner or entity authorized by\n      the copyright owner that is granting the License.\n\n      \"Legal Entity\" shall mean the union of the acting entity and all\n      other entities that control, are controlled by, or are under common\n      control with that entity. For the purposes of this definition,\n      \"control\" means (i) the power, direct or indirect, to cause the\n      direction or management of such entity, whether by contract or\n      otherwise, or (ii) ownership of fifty percent (50%) or more of the\n      outstanding shares, or (iii) beneficial ownership of such entity.\n\n      \"You\" (or \"Your\") shall mean an individual or Legal Entity\n      exercising permissions granted by this License.\n\n      \"Source\" form shall mean the preferred form for making modifications,\n      including but not limited to software source code, documentation\n      source, and configuration files.\n\n      \"Object\" form shall mean any form resulting from mechanical\n      transformation or translation of a Source form, including but\n      not limited to compiled object code, generated documentation,\n      and conversions to other media types.\n\n      \"Work\" shall mean the work of authorship, whether in Source or\n      Object form, made available under the License, as indicated by a\n      copyright notice that is included in or attached to the work\n      (an example is provided in the Appendix below).\n\n      \"Derivative Works\" shall mean any work, whether in Source or Object\n      form, that is based on (or derived from) the Work and for which the\n      editorial revisions, annotations, elaborations, or other modifications\n      represent, as a whole, an original work of authorship. For the purposes\n      of this License, Derivative Works shall not include works that remain\n      separable from, or merely link (or bind by name) to the interfaces of,\n      the Work and Derivative Works thereof.\n\n      \"Contribution\" shall mean any work of authorship, including\n      the original version of the Work and any modifications or additions\n      to that Work or Derivative Works thereof, that is intentionally\n      submitted to Licensor for inclusion in the Work by the copyright owner\n      or by an individual or Legal Entity authorized to submit on behalf of\n      the copyright owner. For the purposes of this definition, \"submitted\"\n      means any form of electronic, verbal, or written communication sent\n      to the Licensor or its representatives, including but not limited to\n      communication on electronic mailing lists, source code control systems,\n      and issue tracking systems that are managed by, or on behalf of, the\n      Licensor for the purpose of discussing and improving the Work, but\n      excluding communication that is conspicuously marked or otherwise\n      designated in writing by the copyright owner as \"Not a Contribution.\"\n\n      \"Contributor\" shall mean Licensor and any individual or Legal Entity\n      on behalf of whom a Contribution has been received by Licensor and\n      subsequently incorporated within the Work.\n\n   2. Grant of Copyright License. Subject to the terms and conditions of\n      this License, each Contributor hereby grants to You a perpetual,\n      worldwide, non-exclusive, no-charge, royalty-free, irrevocable\n      copyright license to reproduce, prepare Derivative Works of,\n      publicly display, publicly perform, sublicense, and distribute the\n      Work and such Derivative Works in Source or Object form.\n\n   3. Grant of Patent License. Subject to the terms and conditions of\n      this License, each Contributor hereby grants to You a perpetual,\n      worldwide, non-exclusive, no-charge, royalty-free, irrevocable\n      (except as stated in this section) patent license to make, have made,\n      use, offer to sell, sell, import, and otherwise transfer the Work,\n      where such license applies only to those patent claims licensable\n      by such Contributor that are necessarily infringed by their\n      Contribution(s) alone or by combination of their Contribution(s)\n      with the Work to which such Contribution(s) was submitted. If You\n      institute patent litigation against any entity (including a\n      cross-claim or counterclaim in a lawsuit) alleging that the Work\n      or a Contribution incorporated within the Work constitutes direct\n      or contributory patent infringement, then any patent licenses\n      granted to You under this License for that Work shall terminate\n      as of the date such litigation is filed.\n\n   4. Redistribution. You may reproduce and distribute copies of the\n      Work or Derivative Works thereof in any medium, with or without\n      modifications, and in Source or Object form, provided that You\n      meet the following conditions:\n\n      (a) You must give any other recipients of the Work or\n          Derivative Works a copy of this License; and\n\n      (b) You must cause any modified files to carry prominent notices\n          stating that You changed the files; and\n\n      (c) You must retain, in the Source form of any Derivative Works\n          that You distribute, all copyright, patent, trademark, and\n          attribution notices from the Source form of the Work,\n          excluding those notices that do not pertain to any part of\n          the Derivative Works; and\n\n      (d) If the Work includes a \"NOTICE\" text file as part of its\n          distribution, then any Derivative Works that You distribute must\n          include a readable copy of the attribution notices contained\n          within such NOTICE file, excluding those notices that do not\n          pertain to any part of the Derivative Works, in at least one\n          of the following places: within a NOTICE text file distributed\n          as part of the Derivative Works; within the Source form or\n          documentation, if provided along with the Derivative Works; or,\n          within a display generated by the Derivative Works, if and\n          wherever such third-party notices normally appear. The contents\n          of the NOTICE file are for informational purposes only and\n          do not modify the License. You may add Your own attribution\n          notices within Derivative Works that You distribute, alongside\n          or as an addendum to the NOTICE text from the Work, provided\n          that such additional attribution notices cannot be construed\n          as modifying the License.\n\n      You may add Your own copyright statement to Your modifications and\n      may provide additional or different license terms and conditions\n      for use, reproduction, or distribution of Your modifications, or\n      for any such Derivative Works as a whole, provided Your use,\n      reproduction, and distribution of the Work otherwise complies with\n      the conditions stated in this License.\n\n   5. Submission of Contributions. Unless You explicitly state otherwise,\n      any Contribution intentionally submitted for inclusion in the Work\n      by You to the Licensor shall be under the terms and conditions of\n      this License, without any additional terms or conditions.\n      Notwithstanding the above, nothing herein shall supersede or modify\n      the terms of any separate license agreement you may have executed\n      with Licensor regarding such Contributions.\n\n   6. Trademarks. This License does not grant permission to use the trade\n      names, trademarks, service marks, or product names of the Licensor,\n      except as required for reasonable and customary use in describing the\n      origin of the Work and reproducing the content of the NOTICE file.\n\n   7. Disclaimer of Warranty. Unless required by applicable law or\n      agreed to in writing, Licensor provides the Work (and each\n      Contributor provides its Contributions) on an \"AS IS\" BASIS,\n      WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or\n      implied, including, without limitation, any warranties or conditions\n      of TITLE, NON-INFRINGEMENT, MERCHANTABILITY, or FITNESS FOR A\n      PARTICULAR PURPOSE. You are solely responsible for determining the\n      appropriateness of using or redistributing the Work and assume any\n      risks associated with Your exercise of permissions under this License.\n\n   8. Limitation of Liability. In no event and under no legal theory,\n      whether in tort (including negligence), contract, or otherwise,\n      unless required by applicable law (such as deliberate and grossly\n      negligent acts) or agreed to in writing, shall any Contributor be\n      liable to You for damages, including any direct, indirect, special,\n      incidental, or consequential damages of any character arising as a\n      result of this License or out of the use or inability to use the\n      Work (including but not limited to damages for loss of goodwill,\n      work stoppage, computer failure or malfunction, or any and all\n      other commercial damages or losses), even if such Contributor\n      has been advised of the possibility of such damages.\n\n   9. Accepting Warranty or Additional Liability. While redistributing\n      the Work or Derivative Works thereof, You may choose to offer,\n      and charge a fee for, acceptance of support, warranty, indemnity,\n      or other liability obligations and/or rights consistent with this\n      License. However, in accepting such obligations, You may act only\n      on Your own behalf and on Your sole responsibility, not on behalf\n      of any other Contributor, and only if You agree to indemnify,\n      defend, and hold each Contributor harmless for any liability\n      incurred by, or claims asserted against, such Contributor by reason\n      of your accepting any such warranty or additional liability.\n\n   END OF TERMS AND CONDITIONS","ref":"license.html"},{"type":"extras","title":"Overview","doc":"# Sourceror 🧙\n\n![Github\nActions](https://github.com/doorgan/sourceror/actions/workflows/main.yml/badge.svg?branch=main)\n[![Coverage\nStatus](https://coveralls.io/repos/github/doorgan/sourceror/badge.svg?branch=main)](https://coveralls.io/github/doorgan/sourceror?branch=main)\n[![Module\nVersion](https://img.shields.io/hexpm/v/sourceror.svg)](https://hex.pm/packages/sourceror)\n[![Hex\nDocs](https://img.shields.io/badge/hex-docs-lightgreen.svg)](https://hexdocs.pm/sourceror/)\n[![Total\nDownload](https://img.shields.io/hexpm/dt/sourceror.svg)](https://hex.pm/packages/sourceror)\n[![License](https://img.shields.io/hexpm/l/sourceror.svg)](https://github.com/doorgan/sourceror/blob/master/LICENSE)\n[![Last\nUpdated](https://img.shields.io/github/last-commit/doorgan/sourceror.svg)](https://github.com/doorgan/sourceror/commits/master)\n\n<!-- MDOC !-->\n\nUtilities to work with Elixir source code.","ref":"readme.html"},{"type":"extras","title":"Installation - Overview","doc":"Add `:sourceror` as a dependency to your project's `mix.exs`:\n\n```elixir\ndefp deps do\n  [\n    {:sourceror, \"~> 1.3\"}\n  ]\nend\n```\n\n#","ref":"readme.html#installation"},{"type":"extras","title":"A note on compatibility - Overview","doc":"Sourceror is compatible with Elixir versions down to 1.10 and OTP 21. For Elixir\nversions prior to 1.13 it uses a vendored version of the Elixir parser and\nformatter modules. This means that for Elixir versions prior to 1.12 it will\nsuccessfully parse the new syntax for stepped ranges instead of raising a\n`SyntaxError`, but everything else should work as expected.","ref":"readme.html#a-note-on-compatibility"},{"type":"extras","title":"Goals of the library - Overview","doc":"- Be as close as possible to the standard Elixir AST.\n- Make working with comments as simple as possible.\n- No dev/prod dependencies, to simplify integration with other tools.","ref":"readme.html#goals-of-the-library"},{"type":"extras","title":"Sourceror's AST - Overview","doc":"Having the AST and comments as separate entities allows Elixir to expose the\ncode formatting utilities without making any changes to it's AST, but also\ndelegates the task of figuring out what's the most appropriate way to work with\nthem to us.\n\nSourceror's take is to use the node metadata to store the comments. This allows\nus to work with an AST that is as close to regular elixir AST as possible. It\nalso allows you to move nodes around without worrying about leaving a comment\nbehind and ending up with misplaced comments.\n\nTwo metadata fields are added to the regular Elixir AST:\n\n- `:leading_comments` - holds the comments directly above the node or are in the\n  same line as it. For example:\n\n  ```elixir\n  test \"parses leading comments\" do\n    quoted = \"\"\"\n    # Comment for :a\n    :a # Also a comment for :a\n    \"\"\" |> Sourceror.parse_string!()\n\n    assert {:__block__, meta, [:a]} = quoted\n    assert meta[:leading_comments] == [\n      %{line: 1, column: 1, previous_eol_count: 1, next_eol_count: 1, text: \"# Comment for :a\"},\n      %{line: 2, column: 4, previous_eol_count: 0, next_eol_count: 1, text: \"# Also a comment for :a\"},\n    ]\n  end\n  ```\n\n- `:trailing_comments` - holds the comments that are inside of the node, but\n  aren't leading any children, for example:\n\n  ```elixir\n  test \"parses trailing comments\" do\n    quoted = \"\"\"\n    def foo() do\n    :ok\n    # A trailing comment\n    end # Not a trailing comment for :foo\n    \"\"\" |> Sourceror.parse_string!()\n\n    assert {:__block__, block_meta, [{:def, meta, _}]} = quoted\n    assert [%{line: 3, text: \"# A trailing comment\"}] = meta[:trailing_comments]\n    assert [%{line: 4, text: \"# Not a trailing comment for :foo\"}] = block_meta[:trailing_comments]\n  end\n  ```\n\nNote that Sourceror considers leading comments to the ones that are found in the\nsame line as a node, and trailing comments to the ones that are found before the\nending line of a node, based on the `end`, `closing` or `end_of_expression`\nline. This also makes the Sourceror AST consistent with the way the Elixir\nformatter works, making it easier to reason about how a given AST would be\nformatted.","ref":"readme.html#sourceror-s-ast"},{"type":"extras","title":"Traversing the AST - Overview","doc":"Elixir provides the `Macro.prewalk`, `Macro.postwalk` and `Macro.traverse`\nfunctions to traverse the AST. You can use the same functions to traverse the\nSourceror AST as well, since it has the same shape as the standard Elixir AST.\n\nSourceror also provides the `Sourceror.prewalk`, `Sourceror.postwalk` and\n`Sourceror.traverse` variants. At the time of writing they are mostly wrappers\naround the standard Elixir functions for AST traversal, but they may be enhanced\nin the future if more AST formats are introduced.\n\nIn addition to these, Sourceror also provides a Zipper implementation for the\nElixir AST. You can learn more about it in the [Zippers\nnotebook](https://hexdocs.pm/sourceror/zippers.html).","ref":"readme.html#traversing-the-ast"},{"type":"extras","title":"Patching the source code - Overview","doc":"You can use Sourceror to manipulate the AST and turn it back into human readable\nElixir code, this is commonly known as writing a \"codemod\". For example, you can\nwrite a codemod to replace calls to `String.to_atom` to\n`String.to_existing_atom`:\n\n```elixir\ntest \"updates the source code\" do\n  source =\n    \"\"\"\n    String.to_atom(foo)\\\n    \"\"\"\n\n  new_source =\n    source\n    |> Sourceror.parse_string!()\n    |> Macro.postwalk(fn\n      {{:., dot_meta, [{:__aliases__, alias_meta, [:String]}, :to_atom]}, call_meta, args} ->\n        {{:., dot_meta, [{:__aliases__, alias_meta, [:String]}, :to_existing_atom]}, call_meta, args}\n\n      quoted ->\n        quoted\n    end)\n    |> Sourceror.to_string()\n\n  assert new_source ==\n    \"\"\"\n    String.to_existing_atom(foo)\\\n    \"\"\"\nend\n```\n\nHowever, this will affect the whole source code, as we are working on the full\nsource AST. Sourceror relies on the Elixir formatter to produce human readable\ncode, so the original code formatting will be lost by using\n`Sourceror.to_string`. If your code is already using the Elixir formatter then\nthis won't be an issue, but it will be an undesirable effect if you're not using\nit.\n\nAn alternative to this is to use Patches instead. A patch is a data structure\nthat specifies the text range that should be replaced, and either a replacement\nstring or a function that takes the text in that range and produces a string\nreplacement.\n\nUsing patches, we could do the same as above, but produce a patch instead of\nmodifying the AST. As a result, only the parts that need to be changed will be\naffected, and the rest of the code keeps the original formatting:\n\n```elixir\ntest \"patches the source code\" do\n  source =\n    \"\"\"\n    case foo do\n      nil ->         :bar\n      _ ->\n\n          String.to_atom(foo)\n\n          end\\\n    \"\"\"\n\n  {_quoted, patches} =\n    source\n    |> Sourceror.parse_string!()\n    |> Macro.postwalk([], fn\n      {{:., dot_meta, [{:__aliases__, alias_meta, [:String]}, :to_atom]}, call_meta, args} = quoted, patches ->\n        range = Sourceror.get_range(quoted)\n        replacement =\n          {{:., dot_meta, [{:__aliases__, alias_meta, [:String]}, :to_existing_atom]}, call_meta, args}\n          |> Sourceror.to_string()\n\n        patch = %{range: range, change: replacement}\n        {quoted, [patch | patches]}\n\n      quoted, patches ->\n        {quoted, patches}\n    end)\n\n  assert Sourceror.patch_string(source, patches) ==\n    \"\"\"\n    case foo do\n      nil ->         :bar\n      _ ->\n\n          String.to_existing_atom(foo)\n\n          end\\\n    \"\"\"\nend\n```\n\nYou have to keep in mind that:\n\n1. If you patch a node that has inner code, like replacing a full `case`, then\n   the contents of the node will be reformatted as well.\n\n2. At the moment, Sourceror won't check for conflicts in the patches ranges, so\n   care needs to be taken to not produce conflicting patches. You may need to do\n   a number of `parse -> patch -> reparse` if you find yourself generating\n   conflicting patches.\n\nSome of the most common patching operations are available in the\n[Sourceror.Patch](https://hexdocs.pm/sourceror/Sourceror.Patch.html) module","ref":"readme.html#patching-the-source-code"},{"type":"extras","title":"Background - Overview","doc":"There have been several attempts at source code manipulation in the Elixir\ncommunity. Thanks to its metaprogramming features, Elixir provides builtin tools\nthat let us get the AST of any Elixir code, but when it comes to turning the AST\nback to code as text, we had limited options. `Macro.to_string/2` is a thing,\nbut the produced code is generally ugly, mostly because of the extra parenthesis\nor because it turns string interpolations into calls to erlang modules, to name\nsome examples. This meant that, even if we could use `Macro.to_string/2` to get\na string and then give that to the Elixir formatter `Code.format_string!/2`, the\noutput would still be suboptimal, as the formatter is not designed to change the\nsemantics of the code, only to pretty print it. For example, call to erlang\nmodules would be kept as is instead of being turned back to interpolations.\n\nWe also had the additional problem of comments being discarded by the tokenizer,\nand literals not having information like line numbers or delimiter characters.\nThis makes the regular AST too lossy to be useful if what we want is to\nmanipulate the source code, because we need as much information as possible to\nbe able to stay as close to the source as possible. There have been several\nproposal in the past to bring all this information to the Elixir AST, but they\nall meant a change that would either break macros due to the addition of new\ntypes of AST nodes, or making a compromise in core Elixir itself by storing\ncomments in the nods metadata. [This\ndiscussion](https://groups.google.com/u/1/g/elixir-lang-core/c/GM0yM5Su1Zc/m/poIKsiEVDQAJ)\nin the Elixir mailing list highlights the various issues faced when deciding if\nand how the comments would be preserved. Arjan Scherpenisse also did a\n[talk](https://www.youtube.com/watch?v=aM0BLWgr0g4&t=117s) where he discusses\nabout the problems of using the standard Elixir AST to build refactoring tools.\n\nDespite of all these issues, the Elixir formatter is still capable of\nmanipulating the source code to pretty print it. Under the hood it does some\nneat tricks to have all this information available: on one hand, it tells the\ntokenizer to extract the comments from the source code and keep it at hand(not\nin the AST itself, but as a separate data structure), and on the other hand it\ntells the parser to wrap literals in block nodes so metadata can be preserved.\nOnce it has all it needs, it can start converting the AST and comments into an\nalgebra document, and ultimately convert that to a string. This functionality\nwas private, and if we wanted to do it ourselves we would have to replicate or\nvendor the Elixir formatter with all its more than 2000 lines of code. This\napproach was explored by Wojtek Mach in\n[wojtekmach/fix](https://github.com/wojtekmach/fix), but it involved vendoring\nthe elixir Formatter code, was tightly coupled to the formatting process, and\nany change in Elixir would break the code.\n\nSince Elixir 1.13 this functionality from the formatter was finally exposed via\nthe `Code.string_to_quoted_with_comments/2` and `Code.quoted_to_algebra/2`\nfunctions. The former gives us access to the list of comments in a shape the\nElixir formatter is able to use, and the latter lets us turn _any arbitrary\nElixir AST_ into an algebra document. If we also give it the list of comments,\nit will merge them together, allowing us to format AST _and_ preserve the\ncomments. Now all we need to care about is of manipulating the AST, and let the\nformatter do the rest.\n\n<!-- MDOC !-->","ref":"readme.html#background"},{"type":"extras","title":"Documentation - Overview","doc":"You can find Sourceror documentation on [Hex\nDocs](https://hexdocs.pm/sourceror/readme.html).","ref":"readme.html#documentation"},{"type":"extras","title":"Examples - Overview","doc":"You can find usage examples in the `examples` folder. You can run them by\ncloning the repo and running `elixir examples/ .exs`.\n\nYou can also find documented examples you can run with\n[Livebook](https://github.com/elixir-nx/livebook) in the `notebooks` folder.","ref":"readme.html#examples"},{"type":"extras","title":"Contributing - Overview","doc":"If you want to contribute to Sourceror, please check our\n[Contributing](https://github.com/doorgan/sourceror/blob/master/CONTRIBUTING.md)\nsection for pointers.","ref":"readme.html#contributing"},{"type":"extras","title":"Getting assistance - Overview","doc":"If you have any questions about Sourceror or need assistance, please open a\nthread in the [Discussions](https://github.com/doorgan/sourceror/discussions)\nsection.","ref":"readme.html#getting-assistance"},{"type":"extras","title":"Copyright and License - Overview","doc":"Copyright (c) 2021 dorgandash@gmail.com\n\nLicensed under the Apache License, Version 2.0 (the \"License\"); you may not use\nthis file except in compliance with the License. You may obtain a copy of the\nLicense at\n[http://www.apache.org/licenses/LICENSE-2.0](http://www.apache.org/licenses/LICENSE-2.0)\n\nUnless required by applicable law or agreed to in writing, software distributed\nunder the License is distributed on an \"AS IS\" BASIS, WITHOUT WARRANTIES OR\nCONDITIONS OF ANY KIND, either express or implied. See the License for the\nspecific language governing permissions and limitations under the License.","ref":"readme.html#copyright-and-license"},{"type":"extras","title":"Zippers","doc":"<!-- vim: set syntax=markdown: -->\n\n<!-- livebook:{\"persist_outputs\":true} -->\n\n# Zippers","ref":"zippers.html"},{"type":"extras","title":"Introduction - Zippers","doc":"Elixir provides the functions `Macro.traverse/4`, `Macro.prewalk/3` and `Macro.postwalk/3` to traverse the AST.\nThey are very useful if we want to collect the nodes of some type or manipulate the *current node* we are visiting, but it becomes increasingly hard to use when we need to remove a node, insert siblings, or get information about other nodes.\nTo mitigate these issues, we can reach for a *zipper*.\n\nA *zipper* is a data structure that represents a *location* in some data, composed of a *focus* (the current element we're visiting) and some *context* that holds the rest of the data from the perspective of that focus.\n\nTo visualize this, we'll use a list of integers from `1` to `10` and say that our current focus is the number `5`:\n\n<!-- livebook:{\"break_markdown\":true} -->\n\n<!-- livebook:{\"force_markdown\":true} -->\n\n```elixir\n[1, 2, 3, 4, 5, 6, 7, 8, 9, 10]\n             ↑\n           focus\n```\n\n<!-- livebook:{\"break_markdown\":true} -->\n\nFrom the perspective of the `5`, the surrounding elements are the numbers from `4` to `1` at it's left,\nand the numbers from `6` to `10` at it's right. We can encode that information in a tuple:\n\n<!-- livebook:{\"break_markdown\":true} -->\n\n<!-- livebook:{\"force_markdown\":true} -->\n\n```elixir\n             focus\n               ↓\n{[4, 3, 2, 1], 5, [6, 7, 8, 9, 10]}\n      ↑                   ↑\nleft siblings      right siblings\n```\n\n<!-- livebook:{\"break_markdown\":true} -->\n\nThat data structure, composed of the focus and the rest of the data as viewed from the focus, is called\na *zipper*. In a traditional list, if we wanted to go to the previous element, we would have to\nstart from the beginning of the list and traverse it all the way down until the current element minus\none position. With a list zipper, however, all we need to do is to pop the first element of the\nleft siblings to get the new focus, and put the old focus in the list of right siblings. After that\nwe get a new zipper focused on `4`:\n\n<!-- livebook:{\"break_markdown\":true} -->\n\n<!-- livebook:{\"force_markdown\":true} -->\n\n```elixir\n{[3, 2, 1], 4, [5, 6, 7, 8, 9, 10]}\n```\n\n<!-- livebook:{\"break_markdown\":true} -->\n\nIf we wanted to move to the right, we would perform the same operation, but popping from the right\nand prepending to the left. Such movements are performed in constant time, making zippers an\nattractive option when a lot of movements over a list are required, but things start to get more\ninteresting when we use zippers with trees. When moving across a list, we are constrained to\nhorizontal movements(left and right), but in trees we have an extra dimension to move around(up and\ndown). So we need to encode at least four things: the focus, the left and right siblings, and the\nparent nodes in the tree.\n\nSourceror provides a zipper implementation for the Elixir AST, let's see what it looks like. First,\nwe need to install Sourceror:\n\n```elixir\nMix.install([\n  :sourceror\n])\n```\n\nThe module that implements zippers for the Elixir AST is `Sourceror.Zipper`.\nWe'll use the alias `Z` for convenience.\n\n```elixir\nalias Sourceror.Zipper, as: Z\n```\n\n<!-- livebook:{\"output\":true} -->\n\n```\nSourceror.Zipper\n```\n\nTo create a zipper for a tree, we use `Z.zip/1`. Let's first start with a tree of lists first, since\nthey are easier to visualize than a full blown AST:\n\n```elixir\nzipper = Z.zip([1, [2, 3], 4, [5, [6, 7]], 8])\n```\n\n<!-- livebook:{\"output\":true} -->\n\n```\n#Sourceror.Zipper<\n  #root\n  [1, [2, 3], 4, [5, [6, 7]], 8]\n>\n```\n\nSourceror represents zippers using the `%Sourceror.Zipper{}` struct.\nWhen this zipper is inspected, the `#root` tag means we're at the topmost node in the tree.\nLet's see what happens when we go down a level:\n\n```elixir\nzipper |> Z.down()\n```\n\n<!-- livebook:{\"output\":true} -->\n\n```\n#Sourceror.Zipper<\n  1\n  #...\n>\n```\n\nNow it's getting more interesting!\nThe current `node` of the zipper is `1`, meaning we went down a level and are now focusing on the first element of the list.\nThe `#...` tag after the `1` means that the current node has right siblings.\nIf we move one position to the right, we get a different zipper:\n\n```elixir\nzipper |> Z.down() |> Z.right()\n```\n\n<!-- livebook:{\"output\":true} -->\n\n```\n#Sourceror.Zipper<\n  #...\n  [2, 3]\n  #...\n>\n```\n\nNow we're focusing on `[2, 3]`, which has siblings both to its left (the `1` we saw previously) and to its right.\n\nSourceror defines a less-verbose implementation of the `Inspect` protocol for zippers, but the default format hides some of the implementation details.\nTo better understand the zipper's internals, we'll set a more verbose default format to use going forward. (You can check out the `Sourceror.Zipper.Inspect` docs for more info.)\n\n```elixir\nSourceror.Zipper.Inspect.default_inspect_as(:raw)\n\nzipper |> Z.down() |> Z.right()\n```\n\n<!-- livebook:{\"output\":true} -->\n\n```\n%Sourceror.Zipper{\n  node: [2, 3],\n  path: %{\n    parent: %Sourceror.Zipper{node: [1, [2, 3], 4, [5, [6, 7]], 8], path: nil},\n    left: [1],\n    right: [4, [5, [6, 7]], 8]\n  }\n}\n```\n\nNow we can see that the zipper struct is composed of two fields, the current `:node` and a `:path` that contains siblings to the left, siblings to the right, and the parent.\n\nBesides moving around, we can also change the tree as a localized operation.\nFor example, let's add an element to the right of our position:\n\n```elixir\nzipper\n|> Z.down()\n|> Z.right()\n|> Z.insert_right(:to_the_right!)\n```\n\n<!-- livebook:{\"output\":true} -->\n\n```\n%Sourceror.Zipper{\n  node: [2, 3],\n  path: %{\n    parent: %Sourceror.Zipper{node: [1, [2, 3], 4, [5, [6, 7]], 8], path: nil},\n    left: [1],\n    right: [:to_the_right!, 4, [5, [6, 7]], 8]\n  }\n}\n```\n\nWe can see that the atom `:to_the_right!` was added to our right siblings.\nLet's now remove it:\n\n```elixir\nzipper\n|> Z.down()\n|> Z.right()\n|> Z.insert_right(:to_the_right!)\n|> Z.right()\n|> Z.remove()\n```\n\n<!-- livebook:{\"output\":true} -->\n\n```\n%Sourceror.Zipper{\n  node: 3,\n  path: %{\n    parent: %Sourceror.Zipper{\n      node: [2, 3],\n      path: %{\n        parent: %Sourceror.Zipper{node: [1, [2, 3], 4, [5, [6, 7]], 8], path: nil},\n        left: [1],\n        right: [4, [5, [6, 7]], 8]\n      }\n    },\n    left: [2],\n    right: nil\n  }\n}\n```\n\nA couple interesting things happened.\nFirst, the item was removed and we can no longer see it in the zipper.\nSecond, the focus moved to the right-most element *inside* the `[2, 3]` list.\nWhenever we remove an item, the focus is moved to the previous element in *depth-first* order.\nIt's as if we did a `Macro.postwalk` step, but backwards.\n\nLet's look a what happens if we go higher in the tree:\n\n```elixir\nzipper\n|> Z.down()\n|> Z.right()\n|> Z.insert_right(:to_the_right!)\n|> Z.up()\n```\n\n<!-- livebook:{\"output\":true} -->\n\n```\n%Sourceror.Zipper{node: [1, [2, 3], :to_the_right!, 4, [5, [6, 7]], 8], path: nil}\n```\n\nInstead of removing the added node, we went back up.\nThe list was reconstructed and the `:path` is now `nil`, meaning we reached the top.\nOne way is that when we go down a level, the tree is deconstructed to adjust to the new focus (the zipper is being opened), and when we go up a level the tree is reconstructed (the zipper is being closed).\n\nNow, while it's useful to know how to move up, down, left and right, this only works if we know the\nshape of the tree before hand. In the real world, we never know the shape of the tree, and so we use\nother kinds of movements to traverse it. For zippers, we have `Z.next/1` and `Z.prev/1` to move\none step forward or backwards in depth-first pre-order, and `Z.traverse/2-3` to perform a depth-first\npre-order traversal while calling a function at each location.\n\nTo demonstrate the ability of zippers to look around at the rest of the tree, we will traverse a tree\nof lists and integers, multiplying each integer by the integer at it's right, or removing it if it's\nalready the rightmost sibling or the following element is another list:\n\n```elixir\n[1, 2, 3, [4, 5, 6, [7, 8, 9], 10, 11], 12, 13, 14]\n|> Z.zip()\n|> Z.traverse(fn\n  %Z{node: x} = zipper when is_integer(x) ->\n    with %Z{node: right} when is_integer(right) <- Z.right(zipper) do\n      Z.replace(zipper, x * right)\n    else\n      _ -> Z.remove(zipper)\n    end\n\n  zipper ->\n    zipper\nend)\n|> Z.node()\n```\n\n<!-- livebook:{\"output\":true} -->\n\n```\n[2, 6, [20, 30, ~c\"8H\", 110], 156, 182]\n```\n\nThe traversal function takes a zipper representing the current position, and must return a new zipper.\nWhen we get back to the top of the tree, the second element will be the atom `:end`, signifying that\nthe traversal ended, preventing subsequent calls to `Z.next/1` from going down the tree again. The\n`Z.node/1` function just extracts the current node from the zipper, so we can look at the final result\ninstead of a zipper.\n\nThat wraps up the introduction to zippers. We'll see them in action in the \"Multi alias expansion\"\nlivebook, where we use them to simplify the traversal and transformations.","ref":"zippers.html#introduction"},{"type":"extras","title":"Expand multi-alias syntax","doc":"<!-- vim: set syntax=markdown: -->\n\n# Expand multi-alias syntax","ref":"expand_multi_alias.html"},{"type":"extras","title":"Introduction - Expand multi-alias syntax","doc":"The multi-alias syntax `alias Foo.{Bar, Baz}`, while handy to alias multiple modules at once,\nmakes module uses harder to search for in large code bases, as mentioned by [credo](https://github.com/rrrene/credo)\nwarnings. We can use Sourceror to fix this issue by expanding this syntax into multiple\ncalls to `alias`, each in its own line.\n\nLet's first start by installing Sourceror:\n\n```elixir\nMix.install([\n  :sourceror\n])\n```\n\nAnd now lets parse an example to get an idea of the kind of structure we're going to work with:\n\n```elixir\nsource = ~S\"\"\"\ndefmodule Foo do\n  alias Foo.{Bar, Baz}\nend\n\"\"\"\n\nquoted = Sourceror.parse_string!(source)\n```\n\nWe need to traverse the ast to find any occurrence of a qualified tuple call(ie.\ncalls with the form `{:., meta, [left, :{}]}` as their first element) that is an argument\nto an `:alias` call. Then for each module inside of the curly brackets, we need to join\nthe module segments from the left hand side with the ones in the right hand side, and\nfinally put them in a call to `:alias`.\n\nFor the traversal part, we can use `Sourceror.postwalk/2`. Postwalk functions will go down\nthe tree to the deepest node, then to the sibling nodes, then to the parent node, and so\non until the whole tree is traversed. A way to think about it is that it traverses bottom\nto top, or that child nodes are always visited first.\n\nTo convert a single alias into multiple ones, we need to extract the left side of the tuple\nand join it with the elements inside of the tuple. For the left part, we can extract it from\nthe dot call we mentioned earlier. The first argument will always be the left hand side, and\nthe second one the atom `:{}`. The elements inside the tuple are just the arguments of the\nqualified tuple call, ie the outer 3-tuple.\n\nEach of these elements will be an `:__aliases__` call. In such calls, the arguments are the\nsegments of the module as regular atoms, so for example the segments for `Foo.Bar` will be\n`[:Foo, :Bar]`. To create a module alias of the expanded `Foo.{Bar}`, we just need to join\nthe segments and put them in an `:__aliases__` call. Finally, that call needs to be wrapped in\na call to `:alias` to effectively create an `alias` expression.\n\nNow we have a list of `:alias` calls, but the traversal function needs to return an AST node,\nnot just a list(that would be considered a list literal). We can work around this for now by\nwrapping the aliases in a `:__block__` and returning that. In other contexts like macros this\nwould change the semantics of the code and it would not behave as we expect, but since we are\ndoing these manipulations to output code as text, we can afford to do it:\n\n```elixir\ndefmodule AliasExpansion do\n  def expand_aliases(quoted) do\n    Sourceror.postwalk(quoted, fn\n      {:alias, _, [{{:., _, [left, :{}]}, _, right}]}, state ->\n        {_, _, base} = left\n\n        aliases =\n          Enum.map(right, fn {_, _, segments} ->\n            aliased = {:__aliases__, [], base ++ segments}\n            {:alias, [], [aliased]}\n          end)\n\n        {{:__block__, [], aliases}, state}\n\n      quoted, state ->\n        {quoted, state}\n    end)\n  end\nend\n\nAliasExpansion.expand_aliases(quoted)\n|> Sourceror.to_string()\n|> IO.puts()\n```\n\nThis works for now because it was just a single expression in the whole module, but it will break\nas soon as we add more expressions:\n\n```elixir\nsource = ~S\"\"\"\ndefmodule Foo do\n  alias Foo.{Bar, Baz}\n  42\nend\n\"\"\"\n\nSourceror.parse_string!(source)\n|> AliasExpansion.expand_aliases()\n|> Sourceror.to_string()\n|> IO.puts()\n```\n\nBecause we wrapped the aliases in a block, if we add more expressions the expanded aliases will\nbe wrapped in parenthesis. This happens because `do` blocks contents are wrapped in a\n`:__block__` whenever they have more than one expression, and what we are doing in our expansion\nis putting a block inside another block, so the formatter will interpret that as a block\nexpression and will wrap it accordingly:\n\n```elixir\nSourceror.parse_string!(~S\"\"\"\ndef foo do\n  :ok\nend\n\"\"\")\n|> IO.inspect(label: \"Single expression\")\n\nSourceror.parse_string!(~S\"\"\"\ndef foo do\n  42\n  :ok\nend\n\"\"\")\n|> IO.inspect(label: \"Multiple expressions\")\n```\n\nOne way to solve this issue is to mark our aliases block as a block that needs to be unwrapped\nif it's inside another block. When traversing, if we encounter a block, we reduce its arguments\nto unwrap any marked block, essentially \"adding multiple nodes\" to the block:\n\n```elixir\ndefmodule AliasExpansionHandledMultiNodes do\n  def expand_aliases(quoted) do\n    Sourceror.postwalk(quoted, fn\n      {:alias, _, [{{:., _, [_, :{}]}, _, _}]} = quoted, state ->\n        aliases = expand_alias(quoted)\n\n        {{:__block__, [unwrap_me?: true], aliases}, state}\n\n      {:__block__, meta, args}, state ->\n        args = Enum.reduce(args, [], &unwrap_aliases/2)\n\n        {{:__block__, meta, args}, state}\n\n      quoted, state ->\n        {quoted, state}\n    end)\n  end\n\n  defp expand_alias({:alias, _, [{{:., _, [left, :{}]}, _, right}]}) do\n    {_, _, base} = left\n\n    aliases =\n      Enum.map(right, fn {_, _, segments} ->\n        aliased = {:__aliases__, [], base ++ segments}\n        {:alias, [], [aliased]}\n      end)\n\n    aliases\n  end\n\n  defp unwrap_aliases({:__block__, [unwrap_me?: true], aliases}, args) do\n    args ++ aliases\n  end\n\n  defp unwrap_aliases(quoted, args) do\n    args ++ [quoted]\n  end\nend\n\nSourceror.parse_string!(source)\n|> AliasExpansionHandledMultiNodes.expand_aliases()\n|> Sourceror.to_string()\n|> IO.puts()\n```","ref":"expand_multi_alias.html#introduction"},{"type":"extras","title":"Handling comments - Expand multi-alias syntax","doc":"Great! Now that we have addressed this issue, there is one last problem we need to take care of.\nIn our current code we are completely ignoring the nodes metadata. This would be fine in most\ncases if we were working with macros, but it becomes a big issue if we want to turn this AST\ninto formatted text. To avoid adding new types of AST nodes, Sourceror places comments in the\nnodes metadata, so if we discard nodes metadata, we would also be discarding it's associated\ncomments. An important aspect of refactoring tools is that they should be able to preserve as\nmuch data as possible while doing the transformation, so we must take comments into account.\n\nThis is easier to see with an example:\n\n```elixir\n~S\"\"\"\n# Some comment\nalias Foo.{Bar, Baz}\n\"\"\"\n|> Sourceror.parse_string!()\n|> AliasExpansionHandledMultiNodes.expand_aliases()\n|> Sourceror.to_string()\n|> IO.puts()\n```\n\nThis issue is easy to avoid if we always remember to pass the metadata around. In this\nparticular example, due to the way Sourceror merges comments, we only need to preserve the alias\nand the individual `:__aliases__` metadata.\n\nThe other thing to keep is that we're getting rid of the first alias and starting anew with the right\nside segments. But this first alias is the one that holds the leading comments for the first\nexpression, which means that if we discard it, we lose the comments right before the multi alias.\n\nWe can solve this by attaching the leading comments to the first one, and the trailing comments to\nthe last one:\n\n```elixir\ndefmodule AliasExpansionHandledComments do\n  def expand_aliases(quoted) do\n    Sourceror.postwalk(quoted, fn\n      {:alias, _, [{{:., _, [_, :{}]}, _, _}]} = quoted, state ->\n        {aliases, state} = expand_alias(quoted, state)\n\n        {{:__block__, [unwrap_me?: true], aliases}, state}\n\n      {:__block__, meta, args}, state ->\n        args = Enum.reduce(args, [], &unwrap_aliases/2)\n\n        {{:__block__, meta, args}, state}\n\n      quoted, state ->\n        {quoted, state}\n    end)\n  end\n\n  defp unwrap_aliases({:__block__, [unwrap_me?: true], aliases}, args) do\n    args ++ aliases\n  end\n\n  defp unwrap_aliases(quoted, args) do\n    args ++ [quoted]\n  end\n\n  defp expand_alias({:alias, alias_meta, [{{:., _, [left, :{}]}, call_meta, right}]}, state) do\n    {_, _, base_segments} = left\n\n    leading_comments = alias_meta[:leading_comments] || []\n    trailing_comments = call_meta[:trailing_comments] || []\n\n    aliases =\n      right\n      |> Enum.map(&segments_to_alias(base_segments, &1))\n      |> put_leading_comments(leading_comments)\n      |> put_trailing_comments(trailing_comments)\n\n    {aliases, state}\n  end\n\n  defp segments_to_alias(base_segments, {_, meta, segments}) do\n    {:alias, meta, [{:__aliases__, [], base_segments ++ segments}]}\n  end\n\n  defp put_leading_comments([first | rest], comments) do\n    [Sourceror.prepend_comments(first, comments) | rest]\n  end\n\n  defp put_trailing_comments(list, comments) do\n    case List.pop_at(list, -1) do\n      {nil, list} ->\n        list\n\n      {last, list} ->\n        last =\n          {:__block__,\n           [\n             trailing_comments: comments,\n             # End of expression newlines higher than 1 will cause the formatter to add an\n             # additional line break after the node. This is entirely optional and only showcased\n             # here to improve the readability of the output\n             end_of_expression: [newlines: 2]\n           ], [last]}\n\n        list ++ [last]\n    end\n  end\nend\n\n~S\"\"\"\n# Some comment\nalias Foo.{Bar, Baz}\n\"\"\"\n|> Sourceror.parse_string!()\n|> AliasExpansionHandledComments.expand_aliases()\n|> Sourceror.to_string()\n|> IO.puts()\n```\n\nOlder versions of sourceror required you to keep track of line numbers, and in an example like the\none above you'd be required to keep track of how line numbers changed to Sourceror would apply\ncorrections and ensure comments ended up in the correct places. Fortunately, newer Sourceror versions\neliminate this issue so you can focus just on moving nodes around and stop worrying about hacky\nline numbers calculations.\n\nYou can try with more convoluted examples and see that the code above produces the output you would\nexpect:\n\n```elixir\nsource = ~S\"\"\"\ndefmodule Sample do\n  # Some aliases\n  alias Foo.{A, B, C, D, E, F}\n\n  # Hello!\n  alias Bar.{G, H, I,\n\n    # Inner comment!\n    # Inner comment 2!\n    # Inner comment 3!\n    J,\n\n    # Comment for K!\n    K # Comment for K 2!\n\n    # Inner last comment!\n    # Inner last comment 2!\n  } # Not an inner comment\n\n  def foo() do\n    # Some scoped alias\n    alias Baz.{A, B, C}\n\n    # Just return :ok\n    :ok\n\n    # At the end\n  end\n\n  # Comment for :hello\n  :hello\nend\n# End of file!\n\"\"\"\n\nsource\n|> Sourceror.parse_string!()\n|> AliasExpansionHandledComments.expand_aliases()\n|> Sourceror.to_string()\n|> IO.puts()\n```","ref":"expand_multi_alias.html#handling-comments"},{"type":"extras","title":"Simplifying traversals with Sourceror.Zipper - Expand multi-alias syntax","doc":"So far the code works, but there is a detail that can be improved. When we expand the alias, we\nare placing all of them in temporary block with the `:flatten_me?` metadata flag, and then\nwe remove it when going up in the postwalk. While this work, it is clearly a hack and we are\nworking around one of the limitation of \"traditional\" tree traversals.\n\nSourceror provides a Zipper implementation for the Elixir AST that makes it easy to traverse\nthe AST in arbitrary directions and perform local manipulations like adding siblings or\nremoving nodes.\n\nA tree Zipper is a data structure that represents a *location* in a tree, rather than the whole\ntree from top to bottom. It's shape is a 2-tuple, where the first element is the current node\nthe zipper is focused on, and the second element is the context, or in other words, the rest\nof the tree from the perspective of the current node.\n\nWhen we are at the top level of\n\n```elixir\nalias Sourceror.Zipper, as: Z\n\nZ.zip([1, [2, 3], 4])\n```\n\nWe get the tuple `{[1, [2, 3], 4], nil}`, where the first element is the tree and the second\nelement is `nil`, signifying that this is a zipper focusing on the root node of the tree.\n\nTo move around in the tree, we can use `Z.up/1` and `Z.down/1` for vertical movement, `Z.left/1`\nand `Z.right/1` for horizontal movement, and `Z.next/1` and `Z.prev/1` to walk the tree in\na depth-first, pre-order fashion. Let's go down the tree and see what happens:\n\n```elixir\nzipper = Z.zip([1, [2, 3], 4])\n\nzipper |> Z.down()\n```\n\nSince we went down a level, we are now at the node `1`, as seen in the first element of the\ntuple, and the metadata is now a tree with the rest of the tree relative to `1`: `l` holds\nthe sibling nodes at the *left* of the current node, `r` holds the sibling nodes at the *right*\nof the current node, and `ptree` holds a the parent tree as a zipper. In a way, every time we\ngo down a level the zipper rips apart the parent\n\n```elixir\nzipper |> Z.down() |> Z.up()\n```\n\nMost of the time we don't care about the metadata/context of the zipper, and we want to get the\ncurrent node out of it. For that we can use `Z.node/1`:\n\n```elixir\nzipper |> Z.down() |> Z.node()\n```\n\nOr we can use `Z.root/1` to go all the way up to\n\n```elixir\nzipper |> Z.down() |> Z.root()\n```\n\nTo move *horizontally*, `Z.left/1` and `Z.right/1` come to the rescue:\n\n```elixir\nzipper |> Z.down() |> Z.right() |> Z.node()\n```\n\n```elixir\nzipper |> Z.down() |> Z.right() |> Z.left() |> Z.node()\n```\n\nFinally, to walk the tree in a similar way to how we would walk it with `Macro.prewalk`, we can\nuse `Z.next/1` and `Z.prev/1`:\n\n```elixir\nzipper |> Z.next() |> Z.next() |> Z.next() |> Z.node()\n```\n\n```elixir\nzipper |> Z.next() |> Z.next() |> Z.next() |> Z.prev() |> Z.prev() |> Z.node()\n```\n\nWalking a tree is nice, but we also need to change it. For that we have tools like `Z.replace/2`,\n`Z.update/2`, `Z.remove/1`, `Z.insert_child` or `Z.traverse/3`.\n\nWe can see them in action by rewriting our `AliasExpansion` module to use the zipper api:\n\n```elixir\ndefmodule AliasExpansionWithZipper do\n  alias Sourceror.Zipper, as: Z\n\n  def expand_aliases(quoted) do\n    Z.zip(quoted)\n    |> Z.traverse(&expand_alias/1)\n    |> Z.root()\n  end\n\n  defp expand_alias(\n         %Z{node: {:alias, alias_meta, [{{:., _, [left, :{}]}, call_meta, right}]}} = zipper\n       ) do\n    {_, _, base_segments} = left\n\n    leading_comments = alias_meta[:leading_comments] || []\n    trailing_comments = call_meta[:trailing_comments] || []\n\n    right\n    |> Enum.map(&segments_to_alias(base_segments, &1))\n    |> put_leading_comments(leading_comments)\n    |> put_trailing_comments(trailing_comments)\n    |> Enum.reverse()\n    |> Enum.reduce(zipper, &Z.insert_right(&2, &1))\n    |> Z.remove()\n  end\n\n  defp expand_alias(zipper), do: zipper\n\n  defp segments_to_alias(base_segments, {_, meta, segments}) do\n    {:alias, meta, [{:__aliases__, [], base_segments ++ segments}]}\n  end\n\n  defp put_leading_comments([first | rest], comments) do\n    [Sourceror.prepend_comments(first, comments) | rest]\n  end\n\n  defp put_trailing_comments(list, comments) do\n    case List.pop_at(list, -1) do\n      {nil, list} ->\n        list\n\n      {last, list} ->\n        last =\n          {:__block__,\n           [\n             trailing_comments: comments,\n             end_of_expression: [newlines: 2]\n           ], [last]}\n\n        list ++ [last]\n    end\n  end\nend\n\nsource\n|> Sourceror.parse_string!()\n|> AliasExpansionWithZipper.expand_aliases()\n|> Sourceror.to_string()\n|> IO.puts()\n```\n\nThe output is the same. Let's confirm:\n\n```elixir\nwithout_zipper =\n  source\n  |> Sourceror.parse_string!()\n  |> AliasExpansionHandledComments.expand_aliases()\n  |> Sourceror.to_string()\n\nwith_zipper =\n  source\n  |> Sourceror.parse_string!()\n  |> AliasExpansionWithZipper.expand_aliases()\n  |> Sourceror.to_string()\n\nwithout_zipper == with_zipper\n```\n\nThe usage of a zipper greatly simplifies the code.\nWe no longer have to resort to hacks like grouping the aliases in a block with the `:flatten_me?`\nmetadata and wait for the traversal to go up to unwrap it. We just say \"add these nodes after me\"\nand that's it. The only detail is that we need to reverse the aliases so the reducing step inserts\nthem in the correct order, but otherwise the code is much more straightforward: find an alias,\ngenerate the expanded aliases, add them to the tree and remove the old alias.","ref":"expand_multi_alias.html#simplifying-traversals-with-sourceror-zipper"},{"type":"extras","title":"Final words - Expand multi-alias syntax","doc":"We've seen how Sourceror approaches comments, we looked at how to perform a simple but yet\nnon-trivial transformation to the source code, and finally we had a look into the Zipper API\nto make powerful transformations in a simple and straightforward way. With these primites,\nyou are now able to leverage Sourceror to create your own transformations, refactoring scripts\nand maybe even macros.","ref":"expand_multi_alias.html#final-words"}],"content_type":"text/markdown","producer":{"name":"ex_doc","version":[48,46,51,49,46,50]}}